from omfit_classes.utils_base import *
from omfit_classes.utils_base import _available_to_user_math, _available_to_user_util, _available_to_user_plot

import numpy as np
import pickle as pypickle
import pickle
import functools
import ast
import filecmp
import types
import shutil
import copy
import difflib
import re
import pprint
import warnings

__all__ = [
    'hide_ptrn',
    'private_ptrn',
    'comment_ptrn',
    'comment_ptrn_in_brackets',
    'number_ptrn',
    'sortHuman',
    'get_bases',
    'parseBuildLocation',
    'parseLocation',
    'traverseLocation',
    'buildLocation',
    'setLocation',
    'dirbaseLocation',
    'traverse',
    'treeLocation',
    'recursiveUpdate',
    'pretty_diff',
    'prune_mask',
    'dynaLoad',
    'dynaLoadKey',
    'dynaSave',
    'dynaLoader',
    'dynaSaver',
    'SortedDict',
    'OMFITdataset',
    'pickle',
    'size_tree_objects',
]

# Useful patterns
hide_ptrn = re.compile(r'^__.*__$')
private_ptrn = re.compile(r'^_.*[^_]+$')
comment_ptrn = re.compile(r'^__comment.*__$')
comment_ptrn_in_brackets = re.compile(r'''.*\[['"]__comment.*__['"]\].*''')
number_ptrn = re.compile(r"[-+]?\d*\.?\d+[eEdD][-+\d]+|[-+\d]+\.\d+|\d+")

_special1 = []


@_available_to_user_util
def sortHuman(inStr):
    """Sort the given list the way that humans expect"""
    outStr = str(inStr).lower()
    tmp = re.findall(number_ptrn, outStr)
    outStr = re.sub(number_ptrn, '\1', outStr)
    for kn in tmp:
        kn = re.sub(r'[dD]', 'e', kn)
        try:
            outStr = re.subn('\1', format(float(kn), "+040.16f"), outStr, 1)[0]
        except ValueError:
            pass
    outStr = re.sub(r'-', 'm', outStr)
    outStr = re.sub(r'\+', 'p', outStr)
    return outStr


def _insort(a, x, caseInsensitive=True):
    lo = 0
    hi = len(a)
    while lo < hi:
        mid = (lo + hi) // 2
        if caseInsensitive and str(x).lower() < str(a[mid]).lower():
            hi = mid
        elif not caseInsensitive and str(x) < str(a[mid]):
            hi = mid
        else:
            lo = mid + 1
    a.insert(lo, x)


def get_bases(clss, tp=[]):
    "Returns a list of strings describing the dependencies of a class"
    if tp == []:
        tp = [clss.__name__]
    bases = getattr(clss, '__bases__')
    if not len(bases):
        return tp
    else:
        for item in bases:
            tp.append(item.__name__)
            get_bases(item, tp)
    return tp


@_available_to_user_util
def different(a, b, precision=0.0):
    """
    Evaluates if two objects are different

    :param a: first object to compare

    :param b: second object to compare

    :param precision: relative precision to which objects are compared

    :return: integer to indicate equal (0) or different (1)
    """
    if isinstance_str(a, ['OMFITexpression', 'OMFITiterableExpression']) and isinstance_str(
        b, ['OMFITexpression', 'OMFITiterableExpression']
    ):
        if a.expression != b.expression:
            return 1
    elif a.__class__ != b.__class__:
        return 1
    elif hasattr(a, 'filename') and hasattr(b, 'filename'):
        if not os.path.exists(a.filename) or not os.path.exists(b.filename):
            return 1
        elif not filecmp.cmp(a.filename, b.filename):
            return 1
        elif not filecmp.cmp(a.filename, b.filename, shallow=False):
            return 1
        elif os.path.split(a.filename)[1] != os.path.split(b.filename)[1]:
            return 1
        return 0
    else:
        try:
            np.testing.assert_equal(a, b)
            return 0
        except Exception:
            if precision == 0.0:
                return 1
            try:
                np.testing.assert_allclose(a, b, rtol=precision)
                return 0
            except Exception:
                return 1
    return 0


@_available_to_user_util
def parseLocation(inv):
    """
    Parse string representation of the dictionary path and return list including root name
    This function can parse things like: OMFIT['asd'].attributes[u'aiy' ]["[  'bla']['asa']"][3][1:5]

    :param inv: string representation of the dictionary path

    :return: list of dictionary keys including rootname
    """
    # look for matching dictionary blocks with matching quotes
    inv = inv.strip()
    quote = False
    starts_at = 0
    splits = []
    k = 0
    char = None
    head = ['']
    while k < len(inv):
        if inv[k] not in ' \t':
            new_char = inv[k]
        if new_char == '[' and not quote:
            if inv[k + 1] in ["'", '"']:
                quote = inv[k + 1]
            elif inv[k + 1] in 'bru' and inv[k + 2] in ["'", '"']:
                quote = inv[k + 2]
            else:
                quote = True
            starts_at = k + 1
        if inv[k] == ']' and (quote is True or last_char == quote):
            quote = False
            splits.append(inv[starts_at:k])

        if not quote and inv[k] not in ' \t[]':
            splits = []
            head = [inv[: k + 1]]
        if inv[k] not in ' \t':
            last_char = inv[k]
        k = k + 1

    # if quote was not closed then the parentheses do not match
    if quote:
        raise SyntaxError('Unbalanced parentheses in ' + inv)

    # eval splits
    for k, item in enumerate(splits):
        try:
            splits[k] = ast.literal_eval(item)
        except SyntaxError:
            if ':' in item:
                splits[k] = item
            else:
                raise
    return head + splits


@_available_to_user_util
def buildLocation(inv):
    """
    Assemble list of keys into dictionary path string

    :param inv: list of dictionary keys including rootname

    :return: string representation of the dictionary path
    """
    tmp = inv[0]
    for item in inv[1:]:
        if isinstance(item, str):
            if ':' in item and not re.findall('[a-zA-Z]', item):
                tmp += '[' + item + ']'
            else:
                tmp += '[' + repr(item) + ']'
        else:
            tmp += '[' + repr(item) + ']'
    return tmp


@_available_to_user_util
def dirbaseLocation(location):
    """
    Takes a string or a list of stirings output by parseLocation()
    and returns two strings for convenient setting of dictionary locations

    >> d, b=dirbaseLocation("OMFIT['dir']['base']")
    >> eval(d)[b]
    d = OMFIT['dir']
    b = 'base'

    :param location: string or a list of stirings output by parseLocation()

    :return: two string, the first one with the path leading to the leaf, the second with the name of the leaf
    """
    if not isinstance(location, list):
        location = parseLocation(location)
    return buildLocation(location[:-1]), location[-1]


@_available_to_user_util
def setLocation(location, value, globals=None, locals=None):
    """
    Takes a string or a list of stirings output by parseLocation()
    and set the leaf to the value provided

    :param location:  string or a list of stirings output by parseLocation()

    :param value: value to set the leaf

    :param globals: global namespace for the evaluation of the location

    :param locals: local namespace for the evaluation of the location

    :return: value
    """
    d, b = dirbaseLocation(location)
    eval(d, globals, locals)[b] = value
    return value


def parseBuildLocation(inv):
    """
    DEPRECATED: use `parseLocation` and `buildLocation` functions instead

    Function to handle locations in the OMFIT tree (i.e. python dictionaries)

    :param inv: input location

    :return:

    * if `inv` is a string, then the dictionary path is split and a list is returned (Note that this function strips the root name)

    * if it's a list, then the dictionary path is assembled and a string is returned (Note that this function assumes that the root name is missing)
    """
    if isinstance(inv, str):
        return parseLocation(inv)[1:]
    elif isinstance(inv, list):
        return buildLocation([''] + inv)
    else:
        raise ValueError('parseBuildLocation accepts either a string or a list')


@_available_to_user_util
def traverseLocation(inv):
    """
    returns list of locations to reach input location

    :param inv: string representation of the dictionary path

    :return: list of locations including rootname to reach input location
    """
    tmp = parseLocation(inv)
    return [buildLocation(tmp[:k]) for k in range(1, len(tmp) + 1)]


@_available_to_user_util
def traverse(
    self, string='', level=100, split=True, onlyDict=False, onlyLeaf=False, skipDynaLoad=False, noSubmodules=False, traverse_classes=(dict,)
):
    """
    Returns a string or list of strings describing the path of every entry/subentries in the dictionary

    :param string: string to be appended in front of all entries

    :param level: maximum depth

    :param split: split the output string into a list of strings

    :param onlyDict: return only dictionary entries (can be a tuple of classes)

    :param onlyLeaf: return only non-dictionary entries (can be a tuple of classes)

    :param skipDynaLoad: skip entries that have .dynaLoad==True

    :param noSubmodules: controls whether to traverse submodules or not

    :param traverse_classes: tuple of classes to traverse

    :return: string or list of string
    """
    string_in = string
    string_out = ''
    if isinstance(self, dict):
        keys = list(self.keys())
    elif isinstance(self, (list, tuple)):
        keys = list(range(len(self)))
    else:
        keys = []
    for kid in keys:
        kidName = "[" + repr(kid) + "]"
        string = string_in + kidName
        # skip also expressions when skipDynaLoad
        if skipDynaLoad and isinstance_str(self[kid], ['OMFITexpression', 'OMFITiterableExpression']):
            continue
        # mention this entry according to `onlyDict` and `onlyLeaf` filters
        if (
            (not onlyDict and not onlyLeaf)
            or (onlyDict is True and isinstance(self[kid], traverse_classes))
            or (isinstance(onlyDict, tuple) and isinstance(self[kid], onlyDict))
            or (onlyLeaf is True and not isinstance(self[kid], traverse_classes))
            or (isinstance(onlyLeaf, tuple) and isinstance(self[kid], onlyLeaf))
        ):
            string_out += string + '\n'
        # do not go deeper if skipDynaLoad and the file has not been loaded
        if skipDynaLoad and hasattr(self[kid], 'dynaLoad') and self[kid].dynaLoad:
            continue
        # go deeper
        if noSubmodules:
            from omfit_classes.omfit_base import OMFITmodule
        if (
            isinstance(self[kid], traverse_classes)
            and (not isinstance(onlyDict, tuple) or isinstance(self[kid], onlyDict))
            and level > 0
            and len(self[kid])
            and not (noSubmodules and isinstance(self[kid], OMFITmodule))
        ):
            level -= 1
            string_out += traverse(self[kid], string, level, False, onlyDict, onlyLeaf, skipDynaLoad, noSubmodules, traverse_classes)
            level += 1
    if split:
        return string_out.strip().strip('\n').split('\n') if string_out else []
    else:
        return string_out


def treeLocation(obj, memo=None):
    """
    Identifies location in the OMFIT tree of an OMFIT object

    NOTE: Typical users should not need to use this function as part of their modules.
    If you find yourself using this function in your modules, it is likely that OMFIT
    already provides the functionality that you are looking for in some other way.
    We recommend reaching out the OMFIT developers team to see if there is an easy
    way to get what you want.

    :param obj: object in the OMFIT tree

    :param memo: used internally to avoid infinite recursions

    :return: string with tree location
    """
    if hasattr(obj, '_OMFITcopyOf') and obj._OMFITcopyOf is not None:
        obj = obj._OMFITcopyOf()

    _nil = []
    if memo is None:
        memo = {}
    y = memo.get(id(obj), _nil)
    if y is not _nil:
        return y

    if not hasattr(obj, '_OMFITparent'):
        try:
            obj._OMFITkeyName = ''
            obj._OMFITparent = None
        except AttributeError:
            # this is for objects which do not accept ._OMFITparent (e.g. int, None, float,...)
            # these are only leafs in the tree and do not need their location in the tree
            # to function anyways.
            return None

    if obj._OMFITparent is None:
        # this is when to treat the head node
        tmp = [obj._OMFITkeyName]
    else:
        # this is for all of the middle nodes
        tmp = treeLocation(obj._OMFITparent)
        if tmp is not None:
            tmp.append(tmp[-1] + obj._OMFITkeyName)

    memo[id(obj)] = tmp
    _keep_alive(obj, memo)
    return tmp


def _keep_alive(x, memo):
    """Keeps a reference to the object x in the memo.

    Because we remember objects by their id, we have
    to assure that possibly temporary objects are kept
    alive by referencing them.
    We store a reference at the id of the memo, which should
    normally not be used unless someone tries to deepcopy
    the memo itself...
    """
    try:
        memo[id(memo)].append(x)
    except KeyError:
        # aha, this is the first one :-)
        memo[id(memo)] = [x]


@_available_to_user_util
def recursiveUpdate(A, B, overwrite=True, **kw):
    """
    Recursive update of dictionary A based on data from dictionary B

    :param A: dictionary A

    :param B: dictionary B

    :param overwrite: force overwrite of duplicates

    :return: updated dictionary
    """
    # for backward compatibility
    if 'overWrite' in kw:
        overwrite = kw.pop('overWrite')
    if len(kw):
        raise TypeError('recursiveUpdate() got an unexpected keyword argument: ' + str(list(kw.keys())))

    def f_traverse(A, B):
        for kid in list(B.keys()):
            if isinstance(B[kid], dict):
                if kid not in A:
                    A[kid] = copy.deepcopy(B[kid])
                elif isinstance(A[kid], dict) and isinstance(B[kid], dict):
                    f_traverse(A[kid], B[kid])
            else:
                if (overwrite and kid in A) or (kid not in A):
                    A[kid] = copy.deepcopy(B[kid])

    f_traverse(A, B)
    return A


def pretty_diff(d, ptrn={}):
    """
    generate "human readable" dictionary output from SortedDict.diff()
    """
    for k in list(d[0].keys()):
        if isinstance(d[0][k][0], dict):
            ptrn[k] = SortedDict()
            pretty_diff(d[0][k], ptrn=ptrn[k])
        else:
            ptrn[k] = d[0][k][0]
    return ptrn


def prune_mask(what, ptrn):
    """
    prune dictionary structure based on mask
    The mask can be in the form of of a `pretty_diff` dictionary or a list of `traverse` strings
    """
    if isinstance(ptrn, dict):
        for k in list(what.keys()):
            if k not in list(ptrn.keys()):
                del what[k]
            elif isinstance(what[k], dict) and isinstance(ptrn[k], dict):
                prune_mask(what[k], ptrn[k])
        return what

    elif isinstance(ptrn, (tuple, list)):

        ptrn = list(ptrn)

        # disregard non-existent paths
        for k, item in list(enumerate(ptrn))[::-1]:
            try:
                eval('what' + item)
            except Exception:
                ptrn.pop(k)

        # expand subtrees
        for item in list(ptrn):
            if isinstance(eval('what' + item), SortedDict):
                ptrn.extend([item + x for x in eval('what' + item).traverse()])

        # expand add parents
        ptrn = set(ptrn)
        for item in list(ptrn):
            rootName = []
            for level in parseLocation(item)[:-1]:
                rootName = rootName + [level]
                ptrn.add(buildLocation(rootName))

        # do the pruning
        for item in what.traverse():
            if item not in ptrn:
                try:
                    exec('del what' + item, globals(), locals())
                except Exception:
                    pass
        return what

    else:
        raise Exception('prune_mask: only list/tuple/dict supported')


def size_tree_objects(location):
    """
    Returns file sizes of objects in the dictionary based on the size of their filename attribute

    :param location: string of the tree location to be analyzed

    :return: dictionary with locations sorted by size
    """
    tmp = traverse(eval(location), onlyDict=False, skipDynaLoad=True)
    sizes = {}
    for item in tmp:
        if hasattr(eval(location + item), 'filename'):
            try:
                obj = eval(location + item)
                if not obj.filename:
                    continue
                size = os.stat(obj.filename).st_size
                if size not in sizes:
                    sizes[size] = []
                sizes[size].append(location + item)
            except Exception as _excp:
                printe('Error sizing object %s : %s' % (location + item, repr(_excp)))
    return sizes


def dynaLoad(f):
    """
    Decorator which calls `dynaLoader` method

    :param f: function to decorate

    :return: decorated function
    """

    @functools.wraps(f)
    def dynamicLoading(self, *args, **kw):
        dynaLoader(self, f)
        return f(self, *args, **kw)

    return dynamicLoading


def dynaLoadKey(f):
    """
    Decorator which calls `dynaLoad` method only if key is not found

    :param f: function to decorate

    :return: decorated function
    """

    @functools.wraps(f)
    def dynamicLoading(self, *args, **kw):
        if args[0] not in self.keyOrder:
            dynaLoader(self, f)
        return f(self, *args, **kw)

    return dynamicLoading


def dynaSave(f):
    """
    Decorator which calls `dynaSaver` method

    :param f: function to decorate

    :return: decorated function
    """

    def doNothing():
        pass

    @functools.wraps(f)
    def dynamicSaving(self, *args, **kw):
        if hasattr_no_dynaLoad(self, 'readOnly') and self.readOnly:
            if hasattr_no_dynaLoad(self, 'modifyOriginal'):
                if self.modifyOriginal and os.path.exists(self.filename) and os.path.samefile(self.link, self.filename):
                    return doNothing
                elif hasattr_no_dynaLoad(self, '_save_by_copy'):
                    return self._save_by_copy(**kw)
        if dynaSaver(self):
            return doNothing
        return f(self, *args, **kw)

    return dynamicSaving


def _docFromDict(f):
    """
    Use the same docstring as for dict

    :param f: function to decorate

    :return: decorated function
    """
    try:
        f.__doc__ = getattr(dict, f.__name__).__doc__
    except AttributeError:
        pass
    return f


def dynaLoader(self, f=None):
    """
    Call `load` function if object has `dynaLoad` attribute set to True
    After calling `load` function the `dynaLoad` attribute is set to False
    """
    if OMFITaux['dynaLoad_switch'] and self.dynaLoad:
        self.dynaLoad = False
        if f is None or f.__name__ != 'load':
            try:
                return self.load()
            except Exception as _excp_load:
                # If an error occurs during loading
                # Clear and reset the dynamic load switch to allow re-tries
                # note: errors could occur because user stops the process
                try:
                    self.clear()
                except Exception as _excp_clear:
                    # if clear() fails, then its exception should be printed but not raised.
                    # the user is interested in the original exception raised by load().
                    printe('The clear() method raised an exception: ' + repr(_excp_clear))
                self.dynaLoad = True
                raise


def dynaSaver(self):
    """
    This function is meant to be called in the .save() function of objects of the class
    `OMFITobject` that support dynamic loading. The idea is that if an object has not
    been loaded, then its file representation has not changed and the original file can be resued.
    This function returns True/False to say if it was successful at saving.
    If True, then the original .save() function can return, otherwise it should go through
    saving the data from memory to file.
    """
    if self.dynaLoad and hasattr_no_dynaLoad(self, 'link') and hasattr_no_dynaLoad(self, 'filename'):
        try:
            printd('Dynamic save: ' + self.filename, level=2, topic='save')
            if os.path.abspath(self.link) != os.path.abspath(self.filename):
                if os.path.exists(self.filename):
                    if filecmp.cmp(self.link, self.filename, shallow=False):
                        self.link = self.filename
                        return True
                    else:
                        os.remove(self.filename)
                try:
                    if OMFITaux.setdefault('hardLinks', False):
                        os.link(self.link, self.filename)
                        printd('Hard link: %s --> %s' % (self.link, self.filename), level=2, topic='save')
                    else:
                        raise Exception('skip')
                except Exception:
                    if os.path.isdir(self.link):
                        shutil.copytree(self.link, self.filename)
                    else:
                        shutil.copy2(self.link, self.filename)
                self.link = self.filename
            return True
        except Exception as _excp:
            printe('Error dynamic save: ' + self.filename + '\n' + repr(_excp))
            return False
    return False


class SortedDict(dict):
    # originally inspired from django/trunk/django/utils/datastructures.py @ 17464
    r"""
    A dictionary that keeps its keys in the order in which they're inserted

    :param data: A dict object or list of (key,value) tuples from which to initialize the new SortedDict object

    :param \**kw: Optional keyword arguments given below

    kw:
        :param caseInsensitive: (bool)  If True, allows self['UPPER'] to yield self['upper'].

        :param sorted: (bool) If True, keep keys sorted alphabetically, instead of by insertion order.

        :param limit: (int) keep only the latest `limit` number of entries (useful for data cashes)

        :param dynaLoad: (bool) Not sure what this does
    """

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        instance._OMFITkeyName = ''
        instance._OMFITparent = None
        instance.keyOrder = []
        instance.caseInsensitive = kwargs.pop('caseInsensitive', False)
        instance.sorted = kwargs.pop('sorted', False)
        instance.limit = kwargs.pop('limit', 0)
        instance.dynaLoad = False
        return instance

    def __init__(self, data=None, *args, **kwargs):
        self.clear()
        self.caseInsensitive = kwargs.pop('caseInsensitive', False)
        self.sorted = kwargs.pop('sorted', False)
        self.limit = kwargs.pop('limit', 0)
        self.dynaLoad = False

        if data is None:
            data = {}
        elif isinstance(data, types.GeneratorType):
            # Unfortunately we need to be able to read a generator twice.  Once
            # to get the data into self with our super().__init__ call and a
            # second time to setup keyOrder correctly
            data = list(data)

        if isinstance(data, dict):
            for key in list(data.keys()):
                self[key] = data[key]
        else:
            for key, value in data:
                self[key] = value

        if self.sorted:
            self.sort()

        self.dynaLoad = kwargs.pop('dynaLoad', False)

        # keep track of what classes have been loaded
        from omfit_classes.utils_base import _loaded_classes

        _loaded_classes.add(self.__class__.__name__)

    def __getattr__(self, attr):
        if attr.startswith('_OMFIT') or attr.startswith('OMFIT'):
            raise AttributeError('bad attribute `%s`' % attr)
        if (
            OMFITaux['dynaLoad_switch']
            and self.dynaLoad
            and attr not in ['__save_kw__', '__tree_repr__', 'modifyOriginal', 'readOnly']
            and 'getattr_infiniteloop_block' not in self.__dict__
        ):
            try:
                if os.environ['USER'] == 'meneghini':
                    printw('%s dynaloading because %s attribute was requested' % (self.__class__.__name__, attr))
                    print_stack()
                self.__dict__['getattr_infiniteloop_block'] = True
                dynaLoader(self)
                return getattr(self, attr)
            finally:
                del self.__dict__['getattr_infiniteloop_block']

        raise AttributeError('bad attribute `%s`' % attr)

    def _setLocation(self, key, value):
        if not hasattr(self, '_OMFITparent') or value is not self._OMFITparent:

            # check if the parent is the OMFIT tree
            inOMFITtree = False
            tmp = self
            while tmp != None and hasattr(tmp, '_OMFITparent'):
                if tmp._OMFITparent is None and tmp._OMFITkeyName == 'OMFIT':
                    inOMFITtree = True
                    break
                tmp = tmp._OMFITparent

            if not (key in self and id(self[key]) == id(value)) and inOMFITtree:
                if isinstance_str(value, ['OMFITexpression', 'OMFITiterableExpression']):
                    value = copy.deepcopy(value)

            try:
                value._OMFITkeyName = "[" + repr(key) + "]"
                value._OMFITparent = self
                value._OMFITcopyOf = None  # this copy goes directly into the tree, so we can set it to None
            except (AttributeError, TypeError):
                pass

        return value

    @_docFromDict
    @dynaLoad
    def __len__(self):
        return super().__len__()

    @_docFromDict
    @dynaLoad
    def __hash__(self):
        return ''.join(list(self.keys())).__hash__()

    @_docFromDict
    @dynaLoad
    def __setitem__(self, key, value):
        key, value = self._checkSetitem(key, value)
        if isinstance_str(key, ['OMFITexpression', 'OMFITiterableExpression']):
            raise ValueError('OMFITexpressions are not valid keys for ' + self.__class__.__name__)
        keyL = self._keyCaseInsensitive(key)
        if keyL != key:
            tmp = self.index(keyL)
            del self[keyL]
            self.keyOrder.insert(tmp, key)
        if key not in self.keyOrder:
            if hasattr(self, 'sorted') and self.sorted:
                _insort(self.keyOrder, key)
            else:
                self.keyOrder.append(key)
        elif self.caseInsensitive:
            self.keyOrder[self.index(key)] = key

        super().__setitem__(key, self._setLocation(key, value))

        # make whatever SortedDict is under a caseInsensitive SortedDict, caseInsensitive itself
        if isinstance(self[key], SortedDict) and self.caseInsensitive:
            self[key].caseInsensitive = self.caseInsensitive

        # if limit>0 limit the number of entries
        while self.limit > 0 and len(self.keyOrder) > self.limit:
            delkey = self.keyOrder[0]
            super().__delitem__(delkey)
            self.keyOrder.remove(delkey)

    def _checkSetitem(self, key, value):
        """
        This method is provided so that subclasses can use it to either:
         1) change the key/value tuple as passed to the __setitem__ method
         2) raise an error because the key-value pair is not acceptable

        :param key: key as passed by the user to the __setitem__ method

        :param value: value as passed by the user to the __setitem__ method

        :return: updated (key, value) tuple
        """
        return key, value

    @_docFromDict
    @dynaLoad
    def __delitem__(self, key):
        key = self._keyCaseInsensitive(key)
        super().__delitem__(key)
        self.keyOrder.remove(key)

    # does not need @dynaLoadKey, because functions that call _keyCaseInsensitive already do
    def _keyCaseInsensitive(self, key):
        if not hasattr(self, 'caseInsensitive'):
            self.caseInsensitive = False
        found = key
        if self.caseInsensitive and key not in self.keyOrder and isinstance(key, str):
            for keyL in self.keyOrder:
                if isinstance(keyL, str) and keyL.lower() == key.lower():
                    found = keyL
                    break
        return found

    @_docFromDict
    @dynaLoadKey
    def __getitem__(self, key):
        key = self._keyCaseInsensitive(key)
        try:
            return super().__getitem__(key)
        except KeyError:
            # if this instance has a fetch method, then call it and try it __getitem__ again
            if hasattr(self, 'fetch'):
                self.fetch()
                return super().__getitem__(key)
            else:
                raise

    @_docFromDict
    @dynaLoadKey
    def __contains__(self, key):
        return super().__contains__(self._keyCaseInsensitive(key))

    def __getstate__(self):
        tmp = copy.copy(self.__dict__)
        for k in list(tmp.keys()):
            if k[:6] == '_OMFIT':
                del tmp[k]
        return tmp, list(self.values())

    def __setstate__(self, tmp):
        if isinstance(tmp, dict):
            # old way of loading sortedDict
            self.__dict__ = tmp
            for key in list(self.keys()):
                self[key] = self._setLocation(key, self[key])
        else:
            # new way of loading sortedDict
            self.limit = False
            self.__dict__.update(tmp[0])
            for key, value in zip(self.keyOrder, tmp[1]):
                self[key] = self._setLocation(key, value)

    @dynaLoad
    def index(self, item):
        """
        returns the index of the item
        """
        return self.keyOrder.index(self._keyCaseInsensitive(item))

    @_docFromDict
    @dynaLoad
    def __iter__(self):
        return iter(self.keyOrder)

    @_docFromDict
    @dynaLoad
    def pop(self, key, *args):
        key = self._keyCaseInsensitive(key)
        result = super().pop(key, *args)
        try:
            self.keyOrder.remove(key)
        except ValueError:
            # Key wasn't in the dictionary in the first place. No problem.
            pass
        return result

    @_docFromDict
    @dynaLoad
    def popitem(self):
        result = super().popitem()
        self.keyOrder.remove(result[0])
        return result

    @_docFromDict
    @dynaLoad
    def items(self):
        return list(zip(self.keyOrder, list(self.values())))

    @_docFromDict
    @dynaLoad
    def iteritems(self):
        for key in self.keyOrder:
            yield key, self[key]

    @dynaLoad
    def keys(self, filter=None, matching=False):
        """
        returns the sorted list of keys in the dictionary

        :param filter: regular expression for filtering keys

        :param matching: boolean to indicate whether to return the keys that match (or not)

        :return: list of keys
        """
        if filter is None:
            return self.keyOrder[:]
        elif not matching:
            return [kkk for kkk in self.keyOrder[:] if not re.match(filter, str(kkk))]
        else:
            return [kkk for kkk in self.keyOrder[:] if re.match(filter, str(kkk))]

    @_docFromDict
    @dynaLoad
    def iterkeys(self):
        return iter(self.keyOrder)

    @_docFromDict
    @dynaLoad
    def values(self):
        return list(map(self.__getitem__, self.keyOrder))

    @_docFromDict
    @dynaLoad
    def itervalues(self):
        for key in self.keyOrder:
            yield self[key]

    @_docFromDict
    @dynaLoad
    def update(self, dict_):
        for key, value in list(dict_.items()):
            self[key] = self._setLocation(key, value)

    @dynaLoad
    def recursiveUpdate(self, other, overwrite=False):
        return recursiveUpdate(self, other, overwrite)

    @dynaLoad
    def setdefault(self, key, default):
        """
        The method setdefault() is similar to get(), but will set dict[key]=default if key is not already in dict

        :param key: key to be accessed

        :param default: default value if key does not exist

        :return: value
        """
        if key not in self:
            if hasattr(self, 'sorted') and self.sorted:
                _insort(self.keyOrder, key)
            else:
                self.keyOrder.append(key)
            self[key] = default
        return self[key]

    @_docFromDict
    @dynaLoad
    def get(self, key, default):
        if key not in self:
            return default
        return self[key]

    @dynaLoad
    def value_for_index(self, index):
        """Returns the value of the item at the given zero-based index"""
        return self[self.keyOrder[index]]

    @dynaLoad
    def insert(self, index, key, value):
        """Inserts the key, value pair before the item with the given index"""
        key = self._keyCaseInsensitive(key)
        if key in self.keyOrder:
            n = self.keyOrder.index(key)
            del self.keyOrder[n]
            if n < index:
                index -= 1
        self.keyOrder.insert(index, key)
        self[key] = value

    @dynaLoad
    def copy(self):
        """Returns a copy of this object"""
        obj = self.__class__(self)
        obj.keyOrder = self.keyOrder[:]
        return obj

    @dynaLoad
    def __repr__(self):
        """returns the keys in their sorted order"""
        return '{%s}' % ', '.join(['%r: %r' % (k, v) for k, v in list(self.items())])

    @_docFromDict
    def clear(self):
        super().clear()
        self.keyOrder = []

    @dynaLoad
    def moveUp(self, index):
        """
        Shift up in key list the item at a given index

        :param index: index to be shifted

        :return: None
        """
        if index < len(self.keyOrder):
            self.keyOrder.insert(index + 1, self.keyOrder.pop(index))

    @dynaLoad
    def moveDown(self, index):
        """
        Shift down in key list the item at a given index

        :param index: index to be shifted

        :return: None
        """
        if index > 0:
            self.keyOrder.insert(index - 1, self.keyOrder.pop(index))

    def __repr__(self):
        return self.__class__.__name__ + '(' + str(list(self.items())) + ')'

    @dynaLoad
    def across(self, what='', sort=False, returnKeys=False):
        """
        Aggregate objects across the tree

        :param what: string with the regular expression to be cut across

        :param sort: sorting of results alphabetically

        :param returnKeys: return keys of elements in addition to objects

        :return: list of objects or tuple with with objects and keys

        >> OMFIT['test']=OMFITtree()
        >> for k in range(5):
        >>   OMFIT['test']['aaa'+str(k)]=OMFITtree()
        >>   OMFIT['test']['aaa'+str(k)]['var']=k
        >> OMFIT['test']['bbb'+str(k)]=-1
        >> print(OMFIT['test'].across("['aaa*']['var']"))
        """
        location = parseBuildLocation(what)
        keys = []
        for k in list(self.keys()):
            if isinstance(k, str):
                if re.match('%r' % location[0], '%r' % k):
                    keys.append(k)
            else:
                if re.match('%r' % location[0], '%r' % repr(k)):
                    keys.append(k)
        if len(location) > 1:
            what = parseBuildLocation(location[1:])
        else:
            what = ''

        if sort:
            index = np.argsort(list(map(float, keys)))
        else:
            index = list(range(len(keys)))

        tmp = []
        for k in index:
            tmp_ = self[keys[k]]
            tmp.append(b2s(eval("tmp_" + what)))
            tmp_
        if returnKeys:
            return tmp, [k for k in np.array(keys)[np.array(index, int)]]
        else:
            return tmp

    @dynaLoad
    def sort(self, key=None, **kw):
        r"""
        :param key: function that returns a string that is used for sorting or dictionary key whose content is used for sorting

        >> tmp=SortedDict()
        >> for k in range(5):
        >>     tmp[k]={}
        >>     tmp[k]['a']=4-k
        >> # by dictionary key
        >> tmp.sort(key='a')
        >> # or equivalently
        >> tmp.sort(key=lambda x:tmp[x]['a'])

        :param \**kw: additional keywords passed to the underlying list sort command

        :return: sorted keys
        """
        if key is None:
            self.keyOrder.sort(key=sortHuman, **kw)

        elif not callable(key):
            self.sort(key=lambda x: self[x][key])

        else:
            self.keyOrder.sort(key=key, **kw)

        return self.keyOrder

    def sort_class(self, class_order=[dict]):
        """
        sort items based on their class

        :param class_order: list containing order of class

        :return: sorted keys
        """
        lst = {}
        for k in self.keyOrder:
            oo = len(class_order)
            for o, c in list(enumerate(class_order)):
                if isinstance(self[k], c):
                    oo = o
                    break
            if hasattr(self[k], '__class__'):
                for o, c in list(enumerate(class_order)):
                    if self[k].__class__.__name__ == c.__name__:
                        oo = o
                        break
            lst.setdefault(oo, []).append(k)
        self.keyOrder = []
        for k in range(len(class_order) + 1):
            if k in lst:
                self.keyOrder += lst[k]
        return self.keyOrder

    @dynaLoad
    def diff(
        self, other, ignoreComments=False, ignoreContent=False, skipClasses=(), noloadClasses=(), precision=0.0, order=True, quiet=True
    ):
        """
        Comparison of a SortedDict

        :param other: other dictionary to compare to

        :param ignoreComments: ignore keys that start and end with "__" (e.g. "__comment__")

        :param ignoreContent: ignore content of the objects

        :param skipClasses: list of class of objects to ignore

        :param noloadClasses: list of class of objects to not load

        :param precision: relative precision to which the comparison is carried out

        :param order: does the order of the keys matter

        :param quiet: verbosity of the comparison

        :return: comparison dictionary
        """
        # todo: should allow taking differences among any type of dictionary, not only sorted dict

        # use difflib.Differ (which operates on strings) to find out differences
        selfEntries = [repr(key) for key in list(self.keys())]
        otherEntries = [repr(key) for key in list(other.keys())]
        tmp = list(difflib.Differ(linejunk=None).compare(selfEntries, otherEntries))
        keys = list(map(ast.literal_eval, [_f for _f in [entry[2:] for entry in tmp if entry[0] != '?'] if _f]))

        # unique keys, keep the ordering, allow caseInsensitive
        tmp = SortedDict(caseInsensitive=self.caseInsensitive)
        for key in keys:
            tmp[key] = ''
        keys = list(tmp.keys())

        if ignoreComments:
            keys = [key for key in keys if not re.match(comment_ptrn, str(key))]

        ndiffs = 0.0
        switch = SortedDict()
        for key in keys:
            if not quiet:
                printi('Compare: ' + str(key))

            if key not in self:
                switch[key] = ['added', False]
                ndiffs += 1.0

            elif key not in other:
                switch[key] = ['removed', False]
                ndiffs += 1.0

            else:
                if isinstance(self[key], skipClasses) or isinstance(other[key], skipClasses):
                    continue

                if isinstance(self[key], noloadClasses) or isinstance(other[key], noloadClasses):
                    if (
                        isinstance(self[key], SortedDict)
                        and isinstance(other[key], SortedDict)
                        and (self[key].dynaLoad or other[key].dynaLoad)
                    ):
                        switch[key] = ['noLoad', False]
                        continue

                if isinstance(self[key], SortedDict) and isinstance(other[key], SortedDict):  # any changes in the subdirs even order
                    tmp = self[key].diff(
                        other[key],
                        ignoreComments=ignoreComments,
                        ignoreContent=ignoreContent,
                        skipClasses=skipClasses,
                        noloadClasses=noloadClasses,
                        precision=precision,
                        quiet=quiet,
                    )
                    if np.sum([len(tmp[0][k]) for k in tmp[0]]) or (
                        order and len([v for k, v in enumerate(self[key]) if list(other[key].keys())[k] != v])
                    ):
                        switch[key] = tmp
                        ndiffs += 1.0

                elif not ignoreContent and different(self[key], other[key], precision=precision):
                    switch[key] = ['changed', False]
                    ndiffs += 1.0

        return [switch, False, keys]

    @dynaLoad
    def pretty_diff(self, other, ignoreComments=False, ignoreContent=False, skipClasses=(), noloadClasses=(), precision=0.0, quiet=True):
        """
        Comparison of a SortedDict in human readable format

        :param other: other dictionary to compare to

        :param ignoreComments: ignore keys that start and end with "__" (e.g. "__comment__")

        :param ignoreContent: ignore content of the objects

        :param skipClasses: list of class of objects to ignore

        :param noloadClasses: list of class of objects to not load

        :param precision: relative precision to which the comparison is carried out

        :param quiet: verbosity of the comparison

        :return: comparison dictionary in pretty to print format
        """
        tmp = self.diff(
            other,
            ignoreComments=ignoreComments,
            ignoreContent=ignoreContent,
            skipClasses=skipClasses,
            noloadClasses=noloadClasses,
            precision=precision,
            quiet=quiet,
        )
        return pretty_diff(tmp)

    @dynaLoad
    def diffKeys(self, other):
        """
        :param other: other dictionary to compare to

        :return: floating point to indicate the ratio of keys that are similar
        """
        # notice that selfKeys and otherKeys are generated with traverse() and do not include comment_ptrn_in_brackets
        selfKeys = set(
            [
                key.lower().split('(')[0]
                for key in traverse(self)
                if isinstance(key, str) and not re.match(comment_ptrn_in_brackets, str(key))
            ]
        )
        otherKeys = set(
            [
                key.lower().split('(')[0]
                for key in traverse(other)
                if isinstance(key, str) and not re.match(comment_ptrn_in_brackets, str(key))
            ]
        )
        if len(selfKeys) < len(otherKeys):
            keys = selfKeys
        else:
            keys = otherKeys

        if len(keys) < 2:
            return 0.0

        ndiffs = 0.0
        for key in keys:
            if len(selfKeys) < len(otherKeys) and key not in otherKeys:
                ndiffs += 1.0
            elif len(selfKeys) >= len(otherKeys) and key not in selfKeys:
                ndiffs += 1.0

        lk = len(keys) * 1.0
        lE = lk - ndiffs
        if lk > 0:
            score = lE * 1.0 / lk
        else:
            score = 1.0
        return score

    @dynaLoad
    def changeKeysCase(self, case=None, recursive=False):
        """
        Change all the keys in the dictionary to be upper/lower case

        :param case: 'upper' or 'lower'

        :param recursive: apply this recursively

        :return: None
        """
        if case is None:
            return
        elif case == 'upper' or case == 'lower':
            for kid in list(self.keys()):
                tmp = self.pop(kid)
                if case == 'upper':
                    self[kid.upper()] = tmp
                elif case == 'lower':
                    self[kid.lower()] = tmp
            for kid in list(self.keys()):
                if recursive and isinstance(self[kid], SortedDict):
                    self[kid].changeKeysCase(case, recursive=True)

    @dynaLoad
    def traverse(self, string='', level=100, onlyDict=False, onlyLeaf=False, skipDynaLoad=False):
        """
        Equivalent to the `tree` command in UNIX

        :param string: prepend this string

        :param level: depth

        :param onlyDict: only subtrees and not the leafs

        :return: list of strings containing the dictionary path to each object
        """
        return traverse(self, string, level, split=True, onlyDict=onlyDict, onlyLeaf=onlyLeaf, skipDynaLoad=skipDynaLoad)

    @dynaLoad
    def walk(self, function, **kw):
        r"""
        Walk every member and call a function on the keyword and value

        :param function: `function(self,kid,**kw)`

        :param \**kw: kw passed to the function

        :return: self
        """
        for kid in list(self.keys()):
            if hasattr(self[kid], 'walk'):
                self[kid].walk(function, **kw)
            else:
                self[kid] = function(self, kid, **kw)
        return self

    @dynaLoad
    def safe_del(self, key):
        """
        Delete key entry only if it is present

        :param key: key to be deleted
        """
        if key in self:
            del self[key]

    @dynaLoad
    def flatten(self):
        """
        The hierarchical structure of the dictionaries is flattened

        :return: SortedDict populated with the flattened content of the dictionary
        """
        tmp = SortedDict(caseInsensitive=self.caseInsensitive)
        for item in list(self.keys()):
            if isinstance(self[item], SortedDict):
                tmp.update(self[item].flatten())
            else:
                tmp[item] = self[item]
        return tmp

    @dynaLoad
    def setFlat(self, key, value):
        """
        recursively searches dictionary for key in order to set a value
        raises KeyError if key could not be found, so this method cannot
        be used to set new entries in the dictionary

        :param key: key to be set

        :param value: value to set
        """
        if key in self:
            self[key] = value
            return
        else:
            for item in list(self.keys()):
                if isinstance(self[item], SortedDict):
                    try:
                        self[item].setFlat(key, value)
                        return
                    except KeyError:
                        pass
        raise KeyError('`%s` could not be found throughout the dictionary' % key)

    @dynaLoad
    def check_location(self, location, value=_special1):
        """
        check if location exist and equals value (if value is specified)

        :param location: location as string

        :param value: value for which to return equal

        :return: True/False

        >> root['SETTINGS'].check_location("['EXPERIMENT']['shot']", 133221)
        """
        try:
            eval('self' + location)
        except KeyError:
            return False
        if value is not _special1 and evalExpr(eval('self' + location)) != evalExpr(value):
            return False
        return True

    def __popup_menu__(self):
        """
        Dummy method to avoid dynamic loading for classes that do not override it

        :return: empty list
        """
        return []

    def todict(self):
        """
        Return a standard Python dictionary representation of the SortedDict

        :return: dictionary
        """
        flat = self.traverse()
        tmp = {}
        for item in flat:
            value = eval(f'self{item}')
            d, b = dirbaseLocation('tmp' + item)
            if isinstance(value, dict):
                eval(d)[b] = {}
            else:
                eval(d)[b] = value
        return tmp


class OMFITdataset(object):
    """
    Subclassing from this class is like subclassing from the xarray.Dataset class
    but without having to deal with the hassle of inheriting from xarrays
    (internally this class uses class composition rather than subclassing).

    Also this class makes it possible to use the OMFIT dynamic loading capabilities.
    All classes that subclass OMFITdataset must define the `.dynaLoad` attribute.

    NOTE: Classes that subclass from OMFITdataset will be identified
    as an xarray.Dataset when using isinstance(..., xarray.Dataset)
    within OMFIT
    """

    def __init__(self, data_vars=None, coords=None, attrs=None):
        r"""
        :param data_vars: see xarray.Dataset

        :param coords: see xarray.Dataset

        :param attrs: see xarray.Dataset
        """
        if self.__class__ == OMFITdataset:
            raise Exception('OMFITdataset is a class that can only be used as a subclass')
        if not hasattr(self, 'dynaLoad'):
            raise Exception(self.__class__.__name__ + ' must set .dynaLoad before calling OMFITdataset.__init__')
        with warnings.catch_warnings(record=True) as w:
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            warnings.filterwarnings('ignore', category=FutureWarning)
            from xarray import Dataset
        self._dataset = Dataset(data_vars=data_vars, coords=coords, attrs=attrs)

    def __getattr__(self, attr):
        # doing this callable test first gets pickling and unpickling to work
        # it used to get stuck in a recursive loop on "dynaLoad" and just testing for attr=='dynaLoad' did not help
        if attr == '__deepcopy__':
            raise AttributeError('Raise not implemented so that we do not use the ._dataset method')
        with warnings.catch_warnings(record=True) as w:
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            warnings.filterwarnings('ignore', category=FutureWarning)
            from xarray import Dataset
        is_callable = callable(getattr(Dataset, attr))
        if self.dynaLoad and (not attr.startswith('_') or attr in ['__getitem__', '__setitem__']):
            self.load()
            self.dynaLoad = False
        return getattr(self._dataset, attr)

    @dynaLoad
    def to_dataset(self):
        """
        Return an xarray.Dataset representation of the data

        :return: xarray.Dataset
        """
        with warnings.catch_warnings(record=True) as w:
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            warnings.filterwarnings('ignore', category=FutureWarning)
            from xarray import Dataset
        return Dataset(self._dataset.data_vars, attrs=self._dataset.attrs)

    @dynaLoad
    def from_dataset(self, dataset):
        """
        Create from xarray.Dataset representation of the data

        """
        with warnings.catch_warnings(record=True) as w:
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            warnings.filterwarnings('ignore', category=FutureWarning)
            from xarray import Dataset

        if isinstance(datatset, Dataset):
            self._dataset = dataset
        else:
            printw("Input is not a proper xarray.Dataset!")

    @dynaLoad
    def __getitem__(self, key):
        # getitem is necessary so that the object is subscriptable
        return self._dataset.__getitem__(key)

    @dynaLoad
    def __setitem__(self, key, value):
        # setitem is necessary so that the object is assignable
        return self._dataset.__setitem__(key, value)

    @dynaLoad
    def __len__(self):
        # len is necessary so that the object is iterable
        return self._dataset.__len__()

    @dynaLoad
    def __iter__(self):
        # iter is necessary so that the object is iterable
        return self._dataset.__iter__()

    @dynaLoad
    def __contains__(self, b):
        # contains is necessary so that the object is iterable
        return self._dataset.__contains__(b)

    def __delitem__(self, b):
        # delete from data set
        del self._dataset[b]


# automatic handle removing of _OMFITxxx attributes when pickling
_dumps = pickle.dumps


@functools.wraps(_dumps)
def _OMFITdumps(x, *args, **kw):
    dynaLoadBkp = OMFITaux['dynaLoad_switch']
    OMFITaux['dynaLoad_switch'] = False
    saveIt = {}
    try:
        if hasattr(x, '__dict__'):
            for k in list(x.__dict__.keys()):
                if k[:6] == '_OMFIT':
                    saveIt[k] = x.__dict__[k]
                    del x.__dict__[k]
        return _dumps(x, *args, **kw)
    finally:
        for k in list(saveIt.keys()):
            x.__dict__[k] = saveIt[k]
        OMFITaux['dynaLoad_switch'] = dynaLoadBkp


pickle.dumps = _OMFITdumps

_dump = pickle.dump


@functools.wraps(_dump)
def _OMFITdump(x, *args, **kw):
    dynaLoadBkp = OMFITaux['dynaLoad_switch']
    OMFITaux['dynaLoad_switch'] = False
    saveIt = {}
    try:
        if hasattr(x, '__dict__'):
            for k in list(x.__dict__.keys()):
                if k[:6] == '_OMFIT':
                    saveIt[k] = x.__dict__[k]
                    del x.__dict__[k]
        return _dump(x, *args, **kw)
    finally:
        for k in list(saveIt.keys()):
            x.__dict__[k] = saveIt[k]
        OMFITaux['dynaLoad_switch'] = dynaLoadBkp


pickle.dump = _OMFITdump

_load = pickle.load


@functools.wraps(_load)
def _OMFITload(*args, **kw):
    dynaLoadBkp = OMFITaux['dynaLoad_switch']
    OMFITaux['dynaLoad_switch'] = False
    try:
        kw.setdefault('encoding', 'latin1')
        return _load(*args, **kw)
    finally:
        OMFITaux['dynaLoad_switch'] = dynaLoadBkp


pickle.load = _OMFITload

_loads = pickle.loads


@functools.wraps(_loads)
def _OMFITloads(*args, **kw):
    dynaLoadBkp = OMFITaux['dynaLoad_switch']
    OMFITaux['dynaLoad_switch'] = False
    try:
        kw.setdefault('encoding', 'latin1')
        return _loads(*args, **kw)
    finally:
        OMFITaux['dynaLoad_switch'] = dynaLoadBkp


pickle.loads = _OMFITloads

if __name__ == "__main__":
    aa = SortedDict({'foo': 1, 'bar': 2, 'yo': 4})
    a = SortedDict({'foo': 1, 'bar': 2, 'yo': 4, 'dd': aa})
    bb = SortedDict({'foo': 0, 'foobar': 3, 'yo': 4})
    b = SortedDict({'foo': 0, 'foobar': 3, 'yo': 4, 'dd': bb})
    diff, switch, keys = a.diff(b)
    print(diff)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

try:
    from omfit_classes.brainfuse import brainfuse
    from omfit_classes.brainfuse import *
except ImportError:
    brainfuse = None
from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.utils_math import *

import random

if brainfuse is not None:

    class OMFITbrainfuse(brainfuse, OMFITascii):
        def __init__(self, filename, **kw):
            OMFITascii.__init__(self, filename, **kw)
            brainfuse.__init__(self, self.filename, **kw)

        def train_net(
            self,
            dB,
            inputNames=[],
            outputNames=[],
            max_iterations=100,
            hidden_layer=None,
            connection_rate=1.0,
            noise=0.0,
            fraction=0.5,
            norm_output={},
            output_mean_0=False,
            robust_stats=0.0,
            weight_decay=0.1,
            spring_decay=1.0,
            activation_function='SIGMOID_SYMMETRIC',
        ):

            """
            :param dB: dictionary of input/output arrays

            :param inputNames: train on these inputs (use all arrays not starting with `OUT_` if not specified)

            :param outputNames: train on these outputs (use all arrays starting with `OUT_` if not specified)

            :param max_iterations: >0 max number of iterations
                                   <0 max number of iterations without improvement

            :param hidden_layer: list of integers defining the NN hidden layer topology

            :param connection_rate: float from 0. to 1. defining the density of the synapses

            :param noise: add gaussian noise to training set

            :param fraction: fraction of data used for training (the rest being for validation)
                             fraction>0 fraction random splitting
                             -1<fraction<0 fraction sequential splitting
                             fraction>1 index sequential splitting

            :param norm_output: normalize outputs

            :param output_mean_0: force average of normalized outputs to have 0 mean

            :param robust_stats: 0<x<100 percentile of data to be considered
                                 =0 mean and std
                                 <0 median and mad

            :param weight_decay: exponential forget of the weight

            :param spring_decay: link training weight decay to the validation error

            :return: std_out of training process
            """

            def f(inv1, inv2):
                try:
                    return abs(np.sum(inv1 - inv2))
                except Exception:
                    return None, None, None

            forbidden = []
            for k1 in sorted(dB.keys()):
                if k1[0] == '_' or np.all(dB[k1] == dB[k1][0]):
                    forbidden.append(k1)
                for k2 in sorted(dB.keys()):
                    if k1 in forbidden or k2 in forbidden:
                        continue
                    v = f(dB[k1], dB[k2])
                    if k1 != k2 and v == 0:
                        forbidden.append(k2)

            if not len(inputNames):
                for k in sorted(dB.keys()):
                    if k not in forbidden and not k.startswith('OUT_') and k[0] != '_':
                        self.inputNames.append(k)
            else:
                self.inputNames = inputNames
            if not len(self.inputNames):
                raise OMFITexception('No inputs!')

            if not len(outputNames):
                for k in sorted(dB.keys()):
                    if k[0] != '_':
                        if np.std(dB[k] - np.mean(dB[k])) != 0:
                            if k.startswith('OUT_') and k[0] != '_':
                                self.outputNames.append(k)
            else:
                self.outputNames = outputNames
            if not len(self.outputNames):
                raise OMFITexception('No outputs!')

            inputs = []
            for k in self.inputNames:
                inputs.append(dB[k].astype(float))
            inputs = np.array(inputs).T

            outputs = []
            for k in self.outputNames:
                outputs.append(dB[k].astype(float))
            outputs = np.array(outputs).T

            def fann_mean(x, dim):
                if robust_stats > 0:
                    tmp = []
                    for k in range(x.shape[-1]):
                        i = np.where(
                            (x[:, k] > np.percentile(x[:, k], robust_stats)) & (x[:, k] < np.percentile(x[:, k], 100 - robust_stats))
                        )[0]
                        if len(i):
                            tmp.append(np.mean(x[i, k]))
                        else:
                            tmp.append(np.mean(x[:, k]))
                elif robust_stats == 0:
                    tmp = np.mean(x, 0)
                else:
                    tmp = np.median(x, 0)
                return tmp

            def fann_std(x, dim):
                if robust_stats > 0:
                    tmp = []
                    for k in range(x.shape[-1]):
                        i = np.where(
                            (x[:, k] > np.percentile(x[:, k], robust_stats)) & (x[:, k] < np.percentile(x[:, k], 100 - robust_stats))
                        )[0]
                        if len(i):
                            tmp.append(np.std(x[i, k]))
                        else:
                            tmp.append(np.std(x[:, k]))
                elif robust_stats == 0:
                    tmp = np.std(x, 0)
                else:
                    tmp = mad(x, 0)

                for k in range(x.shape[-1]):
                    if tmp[k] == 0:
                        tmp[k] = abs(np.mean(x[:, k]))
                    if tmp[k] == 0:
                        tmp[k] = 1.0

                return tmp

            inputs_ = inputs
            outputs_ = outputs

            # normalize
            self.norm_output = []
            for ko, itemo in enumerate(self.outputNames):
                self.norm_output.append([])
                if isinstance(norm_output, dict):
                    for ki, itemi in enumerate(self.inputNames):
                        n0 = 0.0
                        if itemo in norm_output and itemi in norm_output[itemo]:
                            n0 = norm_output[itemo][itemi]
                            print('normalize %s with %s' % (itemo, itemi))
                        self.norm_output[-1].append(n0)
                elif norm_output is True:
                    p, fitfunc = powerlaw_fit(inputs_.T, outputs_[:, ko])
                    print('normalize %s with %s' % (itemo, repr(p)))
                    self.norm_output[-1].extend(p[:-1])
                else:
                    self.norm_output[-1].extend([0.0] * len(self.inputNames))
            outputs_ = self.normOutputs(inputs_, outputs_)

            # scale inputs
            self.scale_mean_in = fann_mean(inputs_, 0)
            self.scale_deviation_in = fann_std(inputs_ - self.scale_mean_in, 0)
            inputs_ = (inputs_ - self.scale_mean_in) / self.scale_deviation_in

            # 0 output mean
            if not output_mean_0:
                self.scale_mean_out = fann_mean(outputs_, 0)
                self.scale_deviation_out = fann_std(outputs_ - self.scale_mean_out, 0)
            else:
                print('output mean 0')
                self.scale_mean_out = fann_mean(outputs_, 0) * 0
                self.scale_deviation_out = np.sqrt(fann_mean(abs(outputs_) ** 2, 0))
            outputs_ = (outputs_ - self.scale_mean_out) / self.scale_deviation_out

            if hidden_layer is None:
                hidden_layer = len(self.inputNames) * 2
            print([len(self.inputNames)] + tolist(hidden_layer) + [len(self.outputNames)])
            self.create_sparse_array(connection_rate, [len(self.inputNames)] + tolist(hidden_layer) + [len(self.outputNames)])
            self.set_activation_function_output(libfann.LINEAR)
            self.set_activation_function_hidden(getattr(libfann, activation_function))
            self.set_training_algorithm(libfann.TRAIN_RPROP)
            self.set_rprop_increase_factor(1.2)
            self.set_rprop_decrease_factor(0.5)
            self.set_train_error_function(libfann.ERRORFUNC_LINEAR)

            self.randomize_weights(-1.0, 1.0)

            if fraction > 0:
                i = np.array(rand(inputs_.shape[0]))
                t_tmpi = inputs_[np.where((i <= fraction))[0], :]
                t_tmpo = outputs_[np.where(i <= fraction)[0], :]
                v_tmpi = inputs_[np.where(i > fraction)[0], :]
                v_tmpo = outputs_[np.where(i > fraction)[0], :]
            elif fraction > -1 and fraction < 0:
                fraction = abs(fraction)
                n = inputs_.shape[0]
                t_tmpi = inputs_[: int(n * fraction), :]
                t_tmpo = outputs_[: int(n * fraction), :]
                v_tmpi = inputs_[int(n * fraction) :, :]
                v_tmpo = outputs_[int(n * fraction) :, :]
            elif abs(fraction) > 1:
                fraction = int(abs(fraction))
                t_tmpi = inputs_[:fraction, :]
                t_tmpo = outputs_[:fraction, :]
                v_tmpi = inputs_[fraction:, :]
                v_tmpo = outputs_[fraction:, :]
            print('training set size: %d' % t_tmpo.shape[0])
            print('validation set size: %d' % v_tmpo.shape[0])

            os.remove(self.filename)

            if noise:
                t_tmpi = t_tmpi * (1 + self.scale_deviation_in[np.newaxis, :] * noise * randn(*t_tmpi.shape))
                t_tmpo = t_tmpo * (1 + self.scale_deviation_out[np.newaxis, :] * noise * randn(*t_tmpo.shape))

            print('Writing train data file')
            txt = [' '.join(map(str, [t_tmpi.shape[0], t_tmpi.shape[1], t_tmpo.shape[1]]))]
            tmp = list(range(t_tmpi.shape[0]))
            random.shuffle(tmp)
            for k in tmp:
                txt.append(' '.join(['%6.6f' % x for x in t_tmpi[k, :]]))
                txt.append(' '.join(['%6.6f' % x for x in t_tmpo[k, :]]))
            trainData = OMFITascii('brainfuse.train', fromString='\n'.join(txt))

            print('Writing validation data file')
            txt = [' '.join(map(str, [v_tmpi.shape[0], v_tmpi.shape[1], v_tmpo.shape[1]]))]
            tmp = list(range(v_tmpi.shape[0]))
            random.shuffle(tmp)
            for k in tmp:
                txt.append(' '.join(['%6.6f' % x for x in v_tmpi[k, :]]))
                txt.append(' '.join(['%6.6f' % x for x in v_tmpo[k, :]]))
            validData = OMFITascii('brainfuse.valid', fromString='\n'.join(txt))

            t0 = time.time()
            self.save()
            inputNames = copy.deepcopy(self.inputNames)
            outputNames = copy.deepcopy(self.outputNames)
            import omfit_classes.OMFITx as OMFITx

            if 'BRAINFUSE_ROOT' not in os.environ:
                raise OMFITexception('Error! BRAINFUSE_ROOT environmental variable is not defined')
            std_out = []
            OMFITx.execute(
                os.environ['BRAINFUSE_ROOT']
                + '/brainfuse_train.exe %s %s %s %d %f %f'
                % (self.filename, trainData.filename, validData.filename, max_iterations, weight_decay, spring_decay),
                std_out=std_out,
            )
            self.load()
            self.inputNames = inputNames
            self.outputNames = outputNames

            data = libfann.training_data()
            data.set_train_data(inputs_, outputs_)
            self.MSE = self.test_data(data)
            data.destroy_train()
            print(self.MSE)

            self.save()

            print('Training timing:' + str(time.time() - t0))
            return std_out


else:
    # associate OMFITbrainfuse to OMFITascii to bring the
    # files along even if they cannot be parsed and used
    class OMFITbrainfuse(OMFITascii):
        pass

    brainfuse = None

    activateNets = None

    activateNetsFile = None

    activateMergeNets = None

__all__ = ['OMFITbrainfuse', 'brainfuse', 'activateNets', 'activateNetsFile', 'activateMergeNets']

############################################
if __name__ == '__main__':
    test_classes_main_header()

    if not os.path.exists('brainfuse.net'):
        print('Unable to test omfit_brainfuse')
        sys.exit()
    OMFITbrainfuse('brainfuse.net')



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.utils_math import is_uncertain
from omfit_classes import namelist

from scipy import signal
import json

# Import both numpy and numpy as np to support evaluation of expressions
import numpy
import numpy as np

__all__ = ['OMFITjson', 'OMFITsettings', 'SettingsName']


def u2s(x, dctType):
    if isinstance(x, str):
        xs = x.encode('utf-8')
        if xs == x:
            return xs
    elif isinstance(x, list):
        return list([u2s(x, dctType) for x in x])
    elif isinstance(x, dict):
        return loader(list(x.items()), dctType)
    return x


# json loader/dumper are borrowed from OMAS
from omas.omas_utils import json_dumper

# customize json_dumper for OMFIT
def dumper(obj):
    if isinstance_str(obj, ['OMFITexpression', 'OMFITiterableExpression']):
        return evalExpr(obj)
    else:
        return json_dumper(obj)


from omas.omas_utils import json_loader

# customize json_loader for OMFIT
def loader(object_pairs, dctType):
    object_pairs = list([(u2s(o[0], dctType), u2s(o[1], dctType)) for o in object_pairs])
    return json_loader(object_pairs, dctType)


class OMFITjson(SortedDict, OMFITascii):
    """
    OMFIT class to read/write json files
    """

    baseDict = SortedDict

    def __init__(self, filename, use_leading_comma=None, **kw):
        r'''
        OMFIT class to parse json files

        :param filename: filename of the json file

        :param use_leading_comma: whether commas whould be leading

        :param \**kw: arguments passed to __init__ of OMFITascii
        '''
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.use_leading_comma = use_leading_comma
        self.dynaLoad = True

    @property
    def use_leading_comma(self):
        return self.OMFITproperties['use_leading_comma']

    @use_leading_comma.setter
    def use_leading_comma(self, use_leading_comma):
        self.OMFITproperties['use_leading_comma'] = use_leading_comma

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            tmp = f.read().strip()

        if not len(tmp):
            return

        self.update(json.loads(tmp, object_pairs_hook=lambda obj: loader(obj, self.baseDict)))

        # figure out if the style of this json is to have a leading comma
        if self.use_leading_comma is None:
            trailing_commas = tmp.count(',\n')
            leading_commas = re.subn('\n(\\s+),', '', tmp)[1]  # https://stackoverflow.com/a/1374893/6605826
            use_leading_comma = leading_commas > trailing_commas
            self.use_leading_comma = leading_commas > trailing_commas

    @dynaSave
    def save(self):
        with open(self.filename, 'w') as f:
            try:
                output = json.dumps(self, default=dumper, indent=1, separators=(',', ': '))
            except AttributeError as _excp:
                raise AttributeError(str(_excp) + ':\nMake sure that the dictionary entries ')
            if self.use_leading_comma:
                output = re.sub(r",\n(\s+)", r"\n\1,", output)
            f.write(output + '\n')

    def __save_kw__(self):
        """
        :return: kw dictionary used to save the attributes to be passed when reloading from OMFITsave.txt
        """
        tmp = self.OMFITproperties.copy()
        if 'use_leading_comma' in tmp and not tmp['use_leading_comma']:
            tmp.pop('use_leading_comma')
        return tmp


class SettingsName(SortedDict):
    """
    Class used for dict-types under OMFITsettings
    """

    baseDict = None

    def __setitem__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, self.baseDict):
            tmp = self.baseDict()
            tmp.update(value)
            value = tmp
        return super().__setitem__(key, value)

    def __repr__(self):
        # we use a simple dictionary representation of Settings
        # to allow nice editing of settings in the OMFIT GUIs
        return self.todict().__repr__()


SettingsName.baseDict = SettingsName


class OMFITsettings(OMFITjson):
    """
    OMFIT class to read/write modules settings
    """

    baseDict = SettingsName

    def __init__(self, filename, **kw):
        super().__init__(filename, **kw)
        self.caseInsensitive = True

    @dynaLoad
    def load(self):
        """
        load as json and if it fails, load as namelist
        """
        try:
            super().load()
        except Exception as _excp:
            tmp = namelist.NamelistFile(self.filename)
            self.update(tmp)

    def __setitem__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, self.baseDict):
            tmp = self.baseDict()
            tmp.update(value)
            value = tmp
        return super().__setitem__(key, value)


def sanitize_all_settings():
    for_all_modules(
        doThis='''
fname = os.path.split(moduleFile)[0] + os.sep + os.path.split(root['SETTINGS'].filename)[1]
root['SETTINGS']=OMFITsettings(fname)
root['SETTINGS'].load()
root['SETTINGS'].use_leading_comma = False
root['SETTINGS'].deploy(fname)
    ''',
        deploy=False,
        skip=False,
    )


############################################
if '__main__' == __name__:
    test_classes_main_header()

    settings = OMFITjson(OMFITsrc + '/../modules/EFIT/SettingsNamelist.txt', use_leading_comma=True)
    print('leading comma', settings.OMFITproperties.get('use_leading_comma', None))



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii

__all__ = ['OMFITspider_bonfit']


class OMFITspider_bonfit(SortedDict, OMFITascii):
    r"""
    OMFIT class used to interface with SPIDER equilibirum bonfit (boundary fit) input file

    :param filename: filename passed to OMFITascii class

    :param \**kw: keyword dictionary passed to OMFITascii class
    """

    def __init__(self, filename, **kw):
        SortedDict.__init__(self)
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.read()
        llines = lines.strip().split('\n')
        llines = [l.split('!')[0] for l in llines]

        ilines = iter(llines)

        def readline(ilines):
            while True:
                tmp = ilines.next().strip()
                if tmp:
                    return tmp

        self['w1'] = np.array(list(map(float, readline(ilines).split())))
        self['w2'] = np.array(list(map(float, readline(ilines).split())))
        n = int(readline(ilines).split()[0])
        self['R'] = []
        self['Z'] = []
        for k in range(n):
            r, z = list(map(float, readline(ilines).split()))
            self['R'].append(r)
            self['Z'].append(z)
        self['R'] = np.array(self['R'])
        self['Z'] = np.array(self['Z'])
        n = int(readline(ilines).split()[0])
        self['control_points'] = {}
        for k in range(n):
            r, z = list(map(float, readline(ilines).split()))
            self['control_points'][k] = np.array([r, z])
        self['xpoint'] = np.array(list(map(float, readline(ilines).split())))
        self['flux'] = float(readline(ilines).split()[0])
        self['xpoint2'] = np.array(list(map(float, readline(ilines).split())))

    @dynaSave
    def save(self):
        txt = []
        for item in self:
            if item == 'control_points':
                txt.append('%d' % len(self[item]))
                for k in self[item]:
                    txt.append(' '.join(list(map(str, self[item][k]))))
            elif item == 'R':
                txt.append('%d' % len(self['R']))
                for r, z in zip(self['R'], self['Z']):
                    txt.append('%s %s' % (r, z))
            elif item == 'Z':
                continue
            elif item == 'flux':
                txt.append(str(self[item]))
            else:
                txt.append(' '.join(list(map(str, self[item]))))
        with open(self.filename, 'w') as f:
            f.write('\n'.join(txt))

    def plot(self, ax=None):
        if ax is None:
            from matplotlib import pyplot

            ax = pyplot.gca()
        ax.plot(self['R'], self['Z'], 'r')
        for cp in self['control_points'].values():
            ax.plot(cp[0], cp[1], 'ro')
        ax.plot(self['xpoint'][0], self['xpoint'][1], '+r', ms=10)
        ax.plot(self['xpoint2'][0], self['xpoint2'][1], '+r', ms=10)
        ax.set_aspect('equal')



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.utils_math import *
from omfit_classes.omfit_base import OMFITtree
from omfit_classes.omfit_osborne import OMFITpFile
from omfit_classes.omfit_eqdsk import OMFITgeqdsk
from omfit_classes.omfit_data import OMFITncDataset, OMFITncDynamicDataset, importDataset
from omas import ODS, omas_environment, cocos_transform, define_cocos
from omfit_classes.omfit_omas_utils import add_generic_OMFIT_info_to_ods

import inspect
import numpy as np
from uncertainties import unumpy, ufloat
from uncertainties.unumpy import std_devs, nominal_values
from scipy import constants

np.seterr(invalid='ignore', divide='ignore')

__all__ = ['OMFITprofiles', 'OMFITprofilesDynamic']

model_tree_species = ['e', '2H1', '4He2', '6Li3', '10B5', '12C6', '14N7', '20Ne10']
# fmt: off
model_tree_quantities = ['angular_momentum', 'angular_momentum_density', 'angular_momentum_density_{species}', 'angular_momentum_{species}',
'dn_{species}_dpsi', 'dT_{species}_dpsi', 'ELM_phase', 'ELM_since_last', 'ELM_until_next', 'fpol', 'jboot_sauter', 'lnLambda',
'n_{species}', 'nclass_sigma', 'nu_star_{species}', 'nu_{species}', 'omega_gyro_{species}_midplane', 'omega_LX_{species}_midplane',
'omega_N_{species}', 'omega_NTV0_{species}', 'omega_P_{species}', 'omega_plasma_{species}', 'omega_RX_{species}_midplane',
'omega_T_{species}', 'omega_tor_{species}', 'omega_tor_{species}_KDG', 'P_brem', 'P_rad', 'P_rad_cNi', 'P_rad_cW', 'P_rad_nNi', 'P_rad_nW',
'P_rad_ZnNi', 'P_rad_ZnW', 'p_tot', 'p_{species}', 'pres', 'psi', 'psi_n', 'q', 'R_midplane', 'resistivity', 'rho', 'T_{species}', 'time',
'Total_Zeff', 'V_pol_{species}_KDG', 'Zavg_Ni', 'Zavg_W', 'Zeff']
# fmt: on


def ugrad1(a2d):
    """
    Gradient along second axis with uncertainty propagation.
    :param a2d: 2D array or uarray
    :return:
    """
    if isinstance(a2d, DataArray):
        a2d = a2d.values
    if is_uncertain(a2d):
        dy = np.gradient(nominal_values(a2d), axis=1)
        ye = std_devs(a2d)
        sigma = np.zeros_like(ye)
        sigma[:, 1:-1] = 0.5 * np.sqrt(ye[:, :-2] ** 2 + ye[:, 2:] ** 2)
        sigma[:, 0] = 0.5 * np.sqrt(ye[:, 0] ** 2 + ye[:, 1] ** 2)
        sigma[:, -1] = 0.5 * np.sqrt(ye[:, -2] ** 2 + ye[:, -1] ** 2)
        result = unumpy.uarray(dy, sigma)
    else:
        result = np.gradient(a2d, axis=1)
    return result


def mZ(species):
    """
    Parse subscript strings and return ion mass and charge

    :param species: subscript strings such as `e`, `12C6`, `2H1`, 'fast_2H1`, ...

    :return: m and Z
    """
    species = str(species).replace('fast_', '')
    if species == 'e':
        Z = -1
        m = constants.m_e
    else:
        m = int(re.sub('([0-9]+)([a-zA-Z]+)([0-9]+)', r'\1', species))
        name = re.sub('([0-9]+)([a-zA-Z]+)([0-9]+)', r'\2', species)
        Z = int(re.sub('([0-9]+)([a-zA-Z]+)([0-9]+)', r'\3', species))
        m *= constants.m_u
    return m, Z


def get_species(derived):
    """
    Identify species and ions that have density information
    """
    species = []
    for key in list(derived.data_vars.keys()):
        if not re.match('^[nT](_fast)?_([0-9]+[a-zA-Z]+[0-9]|e)$', key):
            continue
        s = key.split('_')[-1]
        if '_fast_' in key:
            s = 'fast_' + s
        species.append(s)
    species = tolist(np.unique(species))
    ions = [s for s in species if s not in ['e']]
    ions_with_dens = [i for i in ions if 'n_' + i in derived]
    ions_with_fast = [i.replace('fast_', '') for i in ions if 'fast_' in i]
    return species, ions, ions_with_dens, ions_with_fast


class OMFITprofiles(OMFITncDataset):
    """
    Data class used by OMFITprofiles, CAKE and other
    OMFIT modules for storing experimental profiles data
    """

    def __init__(self, filename, data_vars=None, coords=None, attrs=None, comment=''):
        """
        :param filename: filename of the NetCDF file where data will be saved

        :param data_vars: see xarray.Dataset

        :param coords: see xarray.Dataset

        :param attrs: see xarray.Dataset

        :param comment: String that if set will show in the OMFIT tree GUI
        """
        self.dynaLoad = False
        super().__init__(filename, data_vars=data_vars, coords=coords, attrs=attrs)
        self.OMFITproperties['comment'] = comment

    @property
    def comment(self):
        return self.OMFITproperties['comment']

    @comment.setter
    def comment(self, comment):
        self.OMFITproperties['comment'] = comment

    def __tree_repr__(self):
        if self.comment:
            return self.__class__.__name__ + ': ' + self.comment, []
        else:
            return super().__tree_repr__()

    def to_omas(self, ods=None, times=None):
        """
        :param ods: ODS to which data will be appended

        :return: ods
        """

        if ods is None:
            ods = ODS()
            eq_times = None
        else:
            # Determine if equilibrium is avaliable and for what times.
            eq_times = ods.time('equilibrium') * 1e3  # ODS times are in s, and omfit_profiles are in ms.

        if 'device' in self.attrs:
            ods['dataset_description.data_entry.machine'] = self.attrs['device']
        if 'shot' in self.attrs:
            ods['dataset_description.data_entry.pulse'] = self.attrs['shot']

        # identify fitting coordinate
        for fit_coordinate in ['rho', 'psi_n', None]:
            if fit_coordinate in self.dims:
                break
        if fit_coordinate is None:
            raise ValueError("Fit coordinate should be 'rho' or 'psi_n'")

        # general info
        species, ions, ions_with_dens, ions_with_fast = get_species(self)
        nion = len(ions)
        if times is None:
            times = self['time'].values
        else:
            times = np.atleast_1d(times)

        # figure out time match between eq (if available) and profiles
        if eq_times is None:
            printw("No equilibrium data is avaliable to to_omas(). Some info will not be stored.")
        elif np.all([time in eq_times for time in times]):  # aka is all times have avaliable eq
            printd("Matching equilibrium times found")
        elif np.any([time in eq_times for time in times]):
            printw("Some time slices don't have corresponding equilibria!")
            printw("These time slices will be missing some information.")
        else:
            printw("No equilibrium data is avaliable to to_omas(). Some info will not be stored.")

        # assign both core_profies and edge_profiles but with data from different spatial ranges
        for coredgestring in ['core_profiles', 'edge_profiles']:
            if coredgestring.startswith('core'):
                derived = self.where(self[fit_coordinate] <= 1.0, drop=True)
                core = True
            else:
                derived = self.where(self[fit_coordinate] > 0.8, drop=True)
                core = False

            coredge = ods[coredgestring]
            prop = coredge['ids_properties']
            prop['comment'] = 'Data from OMFITprofiles.to_omas()'
            prop['homogeneous_time'] = True

            coredge['time'] = times / 1e3

            for ti, time in enumerate(times):
                profs = coredge[f'profiles_1d.{ti}']
                profs['time'] = time / 1e3

                # get corresponding eq and extract needed info
                geq = None
                R = None
                Bt = None
                Bp = None  # Make sure we don't accidentally use values from last timeslice
                if eq_times is not None and len(eq_times) > 0 and time in eq_times:
                    i_eq = np.where(time == eq_times)[0][0]  # just the first index
                    geq = OMFITgeqdsk('g0.0').from_omas(ods, time_index=i_eq)
                    # Not the most efficient but can access exiting functions better.
                    if fit_coordinate == 'psi_n':
                        psin = derived['psi_n'].values
                    else:
                        psin = derived['psi_n'].sel(time=time).values
                    psin_eq = np.linspace(0.0, 1.0, len(geq['PRES']))

                    # Re-interpolate
                    R = np.interp(psin, psin_eq, geq['fluxSurfaces']['midplane']['R'])
                    Bt = np.interp(psin, psin_eq, geq['fluxSurfaces']['midplane']['Bt'])
                    Bp = np.interp(psin, psin_eq, geq['fluxSurfaces']['midplane']['Bp'])

                for q in derived.variables:
                    fit = derived[q]
                    if q in ['time']:
                        continue
                    if q == 'rho':
                        if fit_coordinate == 'rho':
                            profs['grid.rho_tor_norm'] = derived['rho'].values
                        else:
                            profs['grid.rho_tor_norm'] = fit.sel(time=time).values
                    elif q == 'psi_n':
                        if fit_coordinate == 'psi_n':
                            profs['grid.rho_pol_norm'] = derived['psi_n'].values
                        else:
                            profs['grid.rho_pol_norm'] = fit.sel(time=time).values
                    elif q == 'n_e':
                        profs['electrons.density_thermal'] = fit.sel(time=time).values
                    elif q == 'T_e':
                        profs['electrons.temperature'] = fit.sel(time=time).values
                    elif q == 'omega_P_e' and core:  # this location is not valid OMAS location for edge profiles
                        profs['electrons.rotation.diamagnetic'] = fit.sel(time=time).values
                    elif '_' in q and q.split('_', 1)[1] in ions:
                        continue

                # thermal ions
                ni = 0
                for ion in ions[::-1]:
                    if ion == 'b' or ion.startswith('fast_'):
                        continue
                    profi = profs[f'ion.{ni}']
                    profi['density_thermal'] = derived[f'n_{ion}'].sel(time=time).values
                    if f'T_{ion}' in derived:
                        profi['temperature'] = derived[f'T_{ion}'].sel(time=time).values
                    ion_details = list(atomic_element(symbol=ion).values())[0]
                    profi['label'] = ion_details['symbol']
                    profi['z_ion'] = float(ion_details['Z_ion'])
                    profi['multiple_states_flag'] = 0
                    profi['element[0].atoms_n'] = 1
                    profi['element[0].z_n'] = float(ion_details['Z'])
                    profi['element[0].a'] = float(ion_details['A'])
                    profi['multiple_states_flag'] = 0
                    if f'V_tor_{ion}' in derived and not (f'omega_tor_{ion}' in derived and 'R_midplane' in derived):
                        profi['velocity.toroidal'] = derived[f'V_tor_{ion}'].sel(time=time).values
                    elif f'omega_tor_{ion}' in derived and 'R_midplane' in derived:
                        profi['velocity.toroidal'] = (
                            derived[f'omega_tor_{ion}'].sel(time=time) * derived['R_midplane'].sel(time=time)
                        ).values
                    if f'V_pol_{ion}' in derived:
                        profi['velocity.poloidal'] = derived[f'V_pol_{ion}'].sel(time=time).values
                    if core:  # extra rotation info for the core profiles. (Not valid nodes for edge)
                        if f'omega_P_{ion}' in derived:
                            profi['rotation.diamagnetic'] = derived[f'omega_P_{ion}'].sel(time=time).values
                        if f'omega_tor_{ion}' in derived:
                            profi['rotation_frequency_tor'] = derived[f'omega_tor_{ion}'].sel(time=time).values
                        if f'V_pol_{ion}' in derived and Bp is not None:
                            # Save to parallel streaming function, this will allow omegp to be calculated from ods
                            profi['rotation.parallel_stream_function'] = derived[f'V_pol_{ion}'].sel(time=time).values / Bp

                    # Advance ni; its important that if it is fast ion and the loop-iteration is skipped, then ni do not advance
                    ni += 1

                # fast ions
                for ion in ions:
                    if ion != 'b' and not ion.startswith('fast_'):
                        continue
                    # Get the 'base' ion for the fast population
                    if ion.startswith('fast_'):
                        base_ion = ion.replace('fast_', '')
                    elif ion == 'b':
                        base_ion == '2H1'  # back compat for '_b' notations
                    ion_details = list(atomic_element(symbol=base_ion).values())[0]

                    # Determin the corresponding ion index
                    ni = len(profs['ion'])
                    for nii in profs['ion']:
                        profi = profs[f'ion.{nii}']
                        if profi['label'] == ion_details['symbol']:
                            ni = nii
                            break
                    profi = profs[f'ion.{ni}']

                    # Add fast_ion data.
                    profi['density_fast'] = derived[f'n_{ion}'].sel(time=time).values
                    if f'p_{ion}' in derived:
                        pfast = derived[f'p_{ion}'].sel(time=time)  #'ion' here would have the form 'fast_2H1' for example
                    else:
                        pfast = derived[f'T_{ion}'].sel(time=time).values * constants.e * derived[f'n_{ion}'].sel(time=time).values
                    profi['pressure_fast_perpendicular'] = (
                        1.0 / 3.0 * pfast
                    )  # Assume isotropic fast ions. Also OMAS treats p_xxx_perp as pressure in one of the perp directions, I think.
                    profi['pressure_fast_parallel'] = 1.0 / 3.0 * pfast

                    # Attach atomic data (from base_ion)
                    profi['label'] = ion_details['symbol']
                    profi['z_ion'] = float(ion_details['Z_ion'])
                    profi['multiple_states_flag'] = 0
                    profi['element[0].atoms_n'] = 1
                    profi['element[0].z_n'] = float(ion_details['Z'])
                    profi['element[0].a'] = float(ion_details['A'])
                    profi['multiple_states_flag'] = 0

                if 'Total_Zeff' in derived:
                    profs['zeff'] = derived['Total_Zeff'].sel(time=time).values

        # Populate total pressure nodes under 'profiles_1d`
        ods.physics_core_profiles_pressures()
        # ods.physics_edge_profiles_pressures() # This function does not exist, but really should.
        return ods

    def model_tree_quantities(self):
        """
        Returns list of MDS+ model_tree_quantities for all species.
        The function will warn if some of the `model_tree_quantities` are missing in OMFIT-source/omfit/omfit_classes/omfit_profiles.py
        and the model tree should be updated

        :return: list of strings
        """
        new_model_tree_quantities = set(model_tree_quantities)
        for item in self.variables:
            match = False
            for s in model_tree_species:
                tmp = item.replace(f'_{s}', '_{species}')
                if tmp != item:
                    match = True
                    break
            if match:
                new_model_tree_quantities.add(tmp)
            else:
                new_model_tree_quantities.add(item)
        new_model_tree_quantities = sorted(list(new_model_tree_quantities), key=lambda x: x.lower())
        if new_model_tree_quantities != model_tree_quantities:
            import textwrap

            printe('WARNING!: Update model_tree_quantities in OMFIT-source/omfit/omfit_classes/omfit_profiles.py')
            printe('WARNING!: and update the OMFIT_PROFS MDS+ model tree')
            printe('-' * 140)
            printe('# fmt: off')
            printe(textwrap.fill(f'model_tree_quantities = {repr(new_model_tree_quantities)}', width=140))
            printe('# fmt: on')
            printe('-' * 140)

        quantities = []
        for item in new_model_tree_quantities:
            if '{' in item:
                for s in model_tree_species:
                    quantities.append(item.format(species=s))
            else:
                quantities.append(item)
        return quantities

    def create_model_tree(self, server, treename='OMFIT_PROFS'):
        """
        Generate MDS+ model tree

        :param server: MDS+ server

        :param treename: MDS+ treename
        """
        from omfit_classes.omfit_mds import OMFITmdsConnection

        conn = OMFITmdsConnection(server)

        quantities = {self.mds_translator(k): None for k in self.model_tree_quantities()}
        quantities['__content__'] = ''
        conn.create_model_tree(treename, '', quantities, clear_subtree=True)

    def to_mds(self, server, shot, treename='OMFIT_PROFS'):
        """
        This script writes the OMFITproflies datase to DIII-D MDS+ and updates d3drdb accordingly

        :param server: MDS+ server

        :param shot: shot to store the data to

        :param treename: MDS+ treename

        :return: runid, treename
        """
        from omfit_classes.omfit_mds import OMFITmdsConnection

        conn = OMFITmdsConnection(server)

        quantities = self.model_tree_quantities()

        # find next available runid in d3drdb for this shot
        from omfit_classes.omfit_rdb import OMFITrdb

        rdb = OMFITrdb(db='code_rundb', server='d3drdb', by_column=True)
        runs = rdb.select(f"select run_id from plasmas where shot={shot} and code_name='OMFITprofiles'", by_column=True)
        for run in range(1, 9999):
            runid = int('%06d%04d' % (shot, run))
            if not len(runs) or runid not in runs['run_id']:
                break
        print(f'Writing OMFITprofiles to MDS+ {runid}')

        # write to MDS+
        quantities = conn.write_dataset(
            treename=treename,
            shot=runid,
            subtree='',
            xarray_dataset=self,
            quantities=quantities,
            translator=lambda x: self.mds_translator(x),
        )
        conn.write(treename=treename, shot=runid, node='__content__', data=';'.join(quantities))

        # add data to d3drdb (after MDS+ so that this is done only if MDS+ write is successful)
        data = {
            'code_name': 'OMFITprofiles',
            'shot': shot,
            'experiment': 'DIII-D',
            'run_type': 'user',
            'tree': treename,
            'start_time': np.min(self['time'].values),
            'stop_time': np.max(self['time'].values),
            'mdsserver': server,
            'run_id': runid,
        }
        rdb.insert('plasmas', data, verbose=False)
        pprint(data)

        return runid, treename

    def mds_translator(self, inv=None):
        """
        Converts strings OMFITprofiles dataset keys to MDS+ nodes less than 12 chars long

        :param inv: tring to which to apply the transformation
                    if `None` the transformation is applied to all of the OMFITprofiles.model_tree_quantities for sanity check

        :return: transformed sting or if inv is None the `mapped_model_2_mds` and `mapped_mds_2_model` dictionaries
        """
        mapper = SortedDict()
        mapper['angular_momentum_density'] = 'mom_dens'
        mapper['angular_momentum'] = 'mom'
        mapper['midplane'] = 'mid'
        mapper['omega_gyro_'] = 'gyrof_'
        mapper['omega_'] = 'w_'
        mapper['2H1'] = 'D'
        mapper['4He2'] = 'He'
        mapper['6Li3'] = 'Li'
        mapper['10B5'] = 'B'
        mapper['12C6'] = 'C'
        mapper['14N7'] = 'N'
        mapper['20Ne10'] = 'Ne'
        mapper['since_last'] = '_last'
        mapper['until_next'] = '_next'

        if inv is not None:
            for match, sub in mapper.items():
                inv = inv.replace(match, sub)
            if len(inv) > 12:
                raise Exception(
                    f'MDS+ OMFITprofiles quantity is longer than 12 chars: {inv}\nUpdate the mds_translator function accordingly'
                )
            return inv
        else:
            model_tree_quantities = self.model_tree_quantities()
            mapped_model_2_mds = SortedDict()
            mapped_mds_2_model = SortedDict()
            for item0 in model_tree_quantities:
                item = item0
                for match, sub in mapper.items():
                    item = item.replace(match, sub)
                if len(item) > 12:
                    raise Exception(f'MDS+ string is longer than 12 chars: {item}')
                if item0 != item and item in model_tree_quantities:
                    raise Exception(f'MDS+ string shadows something else: {item}')
                if item in mapped_mds_2_model:
                    raise Exception(f'Multiple items map to the same quantity: {item0} {mapped_mds_2_model[item]}')
                mapped_model_2_mds[item0] = item
                mapped_mds_2_model[item] = item0
            return mapped_model_2_mds, mapped_mds_2_model

    def to_pFiles(self, eq, times=None, shot=None):
        """
        :param eq: ODS() or dict. (OMFITtree() is a dict)  Needs to contain equilibria information, either in the form
            of the ODS with needed eq already loaded, or as OMFITgeqdsk() objects in the Dict with the time[ms] as
            keys. Times for the eq need to be strict matches to profiles times coord.

        :param times: array like. time for which you would like p files to be generated.

        :param shot: int. shot number, only relevant in generating p file names.

        :return: OMFITtree() containing a series of OMFITpfile objs.
        """

        # Generate times if needed
        if times is None:
            try:
                times = self['time'].values
            except KeyError:
                # Just here to add a helpful hint
                printw("Looking like your OMFITprofiles obj is missing 'time'. Is it properly initialized?")
                raise

        # Get shot if it can be found
        if shot is None:
            if 'shot' in self.attrs:
                shot = self.attrs['shot']
            else:
                shot = 0  # Dummy number for p file names.

        # get good_times
        good_times = []
        if isinstance(eq, ODS):
            # Check times
            good_times = [t for t in times if t in eq.time('equilibrium')]  # ODS().time('...') will throw error if time is inconsistent

        elif isinstance(eq, dict):
            for i, time in enumerate(times):
                d3_time = int(time)
                if d3_time in eq:
                    good_times.append(time)
                else:
                    printw(f"Missing eq data for {time}, it will be skipped!")
        else:
            printw("Input arg 'eq' is in unrecognized format. This will fail!")

        good_times = [t for t in good_times if t in self['time'].values]
        good_times = array(good_times)

        if len(good_times) == 0:
            printw("No valid time found! Each timesilce needs profiles and equilibrium.")
            return
        else:
            printd(f"The following time was found to be good: {good_times}")
            printd("pFiles will be produced for these times only.")

        ods = ODS()
        # Now inject eq info, but only for good_times
        if isinstance(eq, ODS):
            prof_ods.equilibrium.time = np.array(good_times)
            for i, time in enumerate(good_times):
                j = where(eq.times('equilibrium') == time)
                ods[f'equilibrium.time_slice.{i}'] = eq[f'equilibrium.time_slice.{j}']

            prof_ods.physics_consistent_times()  # Generate missing time array
        elif isinstance(eq, dict):
            for i, time in enumerate(good_times):
                d3_time = int(time)
                ods = eq[d3_time].to_omas(ods=ods, time_index=i)
                # This is not very efficient, but does make to_omas() more standardized. Open to change as needs change.
        ods = self.to_omas(times=good_times, ods=ods)

        out_tree = OMFITtree()
        for i, time in enumerate(good_times):
            d3_time = int(time)
            out_tree[d3_time] = OMFITpFile(f'p{shot:06}.{d3_time:05}').from_omas(ods, time_index=i)

        return out_tree

    def get_xkey(self):
        """
        Get the key of the x-coord associated with this data array.

        :returns: str. key of the x-coord associated with the profiles, like 'psi_n'.
        """
        dims = list(self.dims.keys())
        return dims[dims.index('time') - 1]

    def diamagnetic_frequencies(self, spc, update=True):
        """
        Calculate the diamagnetic frequency, and its density / temperature components.

        Formula: \omega_P = -\frac{T}{nZe}\frac{dn}{d\psi} - \frac{1}{Ze}\frac{dT}{d\psi}

        :param spc: Species for which temperature is fit

        :param update: bool. Set to true to update self, if False, only returns and does not update self.
            Gradients, if missing, will always update though.

        """
        freqs = xarray.Dataset()

        m, Z = mZ(spc)

        # calculate gradients if not avaliable
        if f'dn_{spc}_dpsi' not in self and f'n_{spc}' in self:
            dn_dpsi = self.xderiv('n_' + spc, coord='psi')

        if f'dT_{spc}_dpsi' not in self and f'T_{spc}' in self:
            dT_dpsi = self.xderiv('T_' + spc, coord='psi')

        # density part
        if self.check_keys(keys=[f'n_{spc}', f'T_{spc}'] + [f'dn_{spc}_dpsi'], name=f'omega_N_{spc}'):
            omega_N = -self[f'dn_{spc}_dpsi'] * self[f'T_{spc}'] * constants.eV / (self[f'n_{spc}'] * Z * constants.e)
            omega_N.attrs['long_name'] = r"$\omega_N = -\frac{T}{nZe}\frac{dn}{d\psi}$"
            freqs.update({f'omega_N_{spc}': omega_N})
        # temperature part
        if self.check_keys(keys=[f'T_{spc}'] + [f'dT_{spc}_dpsi'], name=f'omega_T_{spc}'):
            omega_T = -self[f'dT_{spc}_dpsi'] * constants.eV / (Z * constants.e)
            omega_T.attrs['long_name'] = r"$\omega_T = -\frac{1}{Ze}\frac{dT}{d\psi}$"
            freqs.update({f'omega_T_{spc}': omega_T})
        # total
        if len(freqs.data_vars) == 2:
            omega_P = omega_N + omega_T
            omega_P.attrs['long_name'] = r"$\omega_{p," + spc + r"} = -\frac{T}{nZe}\frac{dn}{d\psi} -\frac{1}{Ze}\frac{dT}{d\psi}$"
            freqs.update({f'omega_P_{spc}': omega_P})

        if update:
            self.update(freqs)
        return freqs

    def xderiv(self, key, coord='psi', update=True):
        """
        Returns the derivative of the value corresponding to key on the spatial coordinate coord.

        :param key: str. The variable

        :param coord: str. The radial coordinate with respect to which the derivative is taken

        :param update: bool. Set to true to update self, if False, only returns and does not update self.

        :return: Dataset
        """
        result = xarray.Dataset()
        xkey = self.get_xkey()
        if not self.check_keys([key, coord], name=f'd{key}_d{coord}'):
            return result
        dc = np.gradient(np.atleast_2d(self[coord].values), axis=-1)
        dkey = 'd' + key + '_d' + xkey
        if dkey in self and any(isfinite(v) for v in nominal_values(self[dkey].values).flat) and coord != xkey:
            # if we have derivative UQ from the fit method, keep it and just swap coordinates
            dx = np.gradient(np.atleast_2d(self[xkey].values), axis=-1)
            dkdc = DataArray(self[dkey] * (dx / dc), coords=self[key].coords)
        else:
            # otherwise we have to calculate the derivative numerically - huge uncertainties when propagated
            try:
                dkdc = DataArray(ugrad1(self[key].values) / dc, coords=self[key].coords)
            except ZeroDivisionError:  # This happens if dc array have 0s and self[key].values is a uarray/has uncertainty.
                i_bad = dc == 0
                dc[i_bad] = 1
                dkdc_arr = ugrad1(self[key].values) / dc
                dkdc_arr[i_bad] = ufloat(np.nan, np.nan)
                dkdc = DataArray(ugrad1(self[key].values) / dc, coords=self[key].coords)
        if coord in ['rho', 'psi_n']:
            dkdc.attrs['units'] = self[key].attrs.get('units', '')
        dkdc.attrs['long_name'] = r'd{:} / d$\{:}$'.format(key, coord)
        result['d' + key + '_d' + coord] = dkdc

        if update:
            self.update(result)
        return result

    def check_keys(self, keys=[], name='', print_error=True):
        """
        Check to make sure required data is available
        """
        missing = []
        for k in keys:
            if not k in self:
                missing.append(k)
        if missing:
            if not name:
                name = 'value'
            if print_error:
                printw('  WARNING: Could not form {:}. Missing {:}'.format(name, ', '.join(['`%s`' % x for x in missing])))
            return False
        return True


class OMFITprofilesDynamic(OMFITncDynamicDataset):
    """
    Class for dynamic calculation of derived quantities

    :Examples:

    Initialize the class with a filename and FIT Dataset.
    >> tmp=OMFITprofiles('test.nc', fits=root['OUTPUTS']['FIT'], equilibrium=root['OUTPUTS']['SLICE']['EQ'], root['SETTINGS']['EXPERIMENT']['gas'])

    Accessing a quantity will dynamically calculate it.
    >> print tmp['Zeff']

    Quantities are then stored (they are not calculated twice).
    >> tmp=OMFITprofiles('test.nc',
                          fits=root['OUTPUTS']['FIT'],
                          equilibrium=root['OUTPUTS']['SLICE']['EQ'],
                          main_ion='2H1')
    >> uband(tmp['rho'],tmp['n_2H1'])
    """

    def __init__(self, filename, fits=None, equilibrium=None, main_ion='2H1', **kw):
        OMFITncDataset.__init__(self, filename, **kw)

        if fits:
            self.update(fits)

            # profile fit dimension
            dims = list(self.dims.keys())
            xkey = dims[dims.index('time') - 1]

            # collect some meta data about which particle species have what info available
            self.attrs['main_ion'] = str(main_ion)
            species, ions, ions_with_dens, ions_with_fast = get_species(self)
            self['species'] = DataArray(species, dims=['species'])
            ions += [self.attrs['main_ion']]
            self['ions'] = DataArray(ions, dims=['ions'])
            ions_with_dens += [self.attrs['main_ion']]
            self['ions_with_dens'] = DataArray(ions_with_dens, dims=['ions_with_dens'])
            ions_with_rot = [key.replace('omega_tor_', '') for key in self if key.startswith('omega_tor') and key != 'omega_tor_e']
            self['ions_with_rot'] = DataArray(ions_with_rot, dims=['ions_with_rot'])

            # interpolate other radial coordinates from equilibrium
            printd('- Interpolating equilibrium quantities')
            needed = {
                'avg_R',
                'avg_R**2',
                'avg_1/R',
                'avg_1/R**2',
                'avg_Btot',
                'avg_Btot**2',
                'avg_vp',
                'avg_q',
                'avg_fc',
                'avg_F',
                'avg_P',
                'geo_psi',
                'geo_R',
                'geo_Z',
                'geo_a',
                'geo_eps',
                'geo_vol',
            }
            eqcoords = needed.intersection(list(equilibrium.keys()))
            for meas in eqcoords:
                if 'profile_' + xkey != meas and 'profile_psi_n' in equilibrium[meas]:
                    yy = []
                    for t in self['time'].values:
                        eq_t = equilibrium.sel(time=t, method='nearest')
                        x = np.squeeze(eq_t['profile_' + xkey])
                        y = np.squeeze(nominal_values(eq_t[meas].values))
                        yy.append(interp1e(x, y)(self[xkey].values))

                        # Ensure that 'q' is not extrapolated outside the separatrix
                        if meas == 'profile_q':
                            mask = self[xkey].values > 1.0
                            yy[-1][mask] = np.nan
                    yy = np.array(yy)
                    key = meas.replace('profile_', '')
                    self[key] = DataArray(
                        yy, coords=[fits['time'], fits[xkey]], dims=['time', xkey], attrs=copy.deepcopy(equilibrium[meas].attrs)
                    )
            self.set_coords(eqcoords)
            # reassign global attrs clobbered when assigning eq DataArrays with attrs
            self.attrs['main_ion'] = str(main_ion)
            self.save()

        self.update_dynamic_keys(self.__class__)

    def __getitem__(self, key):
        # map specific quantities to class functions
        mapper = {}
        mapper['n_' + self.attrs['main_ion']] = 'calc_n_main_ion'
        mapper['T_' + self.attrs['main_ion']] = 'calc_T_main_ion'

        # resolve mappings
        if key not in self:
            if key in mapper:
                getattr(self, mapper[key])()
                if mapper[key] in self._dynamic_keys:
                    self._dynamic_keys.pop(self._dynamic_keys.index(mapper[key]))

        # return value
        return OMFITncDynamicDataset.__getitem__(self, key)

    def calc_n_main_ion(self):
        """
        Density of the main ion species.
        Assumes quasi-neutrality.

        :return: None. Updates the instance's Dataset in place.

        """
        main_ion = str(self.attrs['main_ion'])
        mg, zg = mZ(main_ion)
        nz = self['n_e']
        for key in self['ions_with_dens'].values:
            if key != main_ion:
                nz -= self['n_' + key].values * mZ(key)[1]
        self['n_' + main_ion] = nz / zg
        invalid = np.where(self['n_' + main_ion].values <= 0)[0]
        if len(invalid) > 0:
            printe('  Had to force main ion density to be always positive!')
            printe('  This will likely present a problem when running transport codes!')
            valid = np.where(self['n_' + main_ion].values > 0)[0]
            self['n_' + main_ion].values[invalid] = np.nanmin(self['n_' + main_ion].values[valid])

    def calc_T_main_ion(self):
        """
        Temperature of the main ion species.
        Assumes it is equal to the measured ion species temperature.
        If there are multiple impurity temperatures measured, it uses the first one.

        :return: None. Updates the instance's Dataset in place.

        """
        main_ion = str(self.attrs['main_ion'])
        impurities_with_temp = [k for k in self['ions'].values if k != 'b' and 'T_' + k in list(self.keys())]
        nwith = len(impurities_with_temp)
        if nwith == 0:
            raise OMFITexception("No main or impurity ion temperatures measured")
        if nwith > 1:
            printw(
                "WARNING: Multiple impurities temperatures measured, setting main ion temperature based on {:}".format(
                    impurities_with_temp[0]
                )
            )
        for ion in impurities_with_temp:
            self['T_' + main_ion] = self[f'T_{ion}'] * 1
            break

    def calc_Zeff(self):
        r"""
        Effective charge of plasma.

        Formula: Z_{eff} = \sum{n_s Z_s^2} / \sum{n_s Z_s}

        :return: None. Updates the instance's Dataset in place.

        """
        # calculate Zeff (not assuming quasi-neutrality)
        nz_sum = np.sum([self['n_' + i].values * mZ(i)[1] for i in self['ions_with_dens'].values], axis=0)
        nz2_sum = np.sum([self['n_' + i].values * mZ(i)[1] ** 2 for i in self['ions_with_dens'].values], axis=0)
        z_eff = nz2_sum / nz_sum + 0 * self['n_e'].rename('Zeff')
        z_eff.attrs['long_name'] = r'$Z_{eff}$'
        self['Zeff'] = z_eff

    def calc_Total_Zeff(self):
        r"""
        Effective charge of plasma.

        Formula: Z_{eff} = \sum{n_s Z_s^2} / \sum{n_s Z_s}

        :return: None. Updates the instance's Dataset in place.

        """
        main_ion = str(self.attrs['main_ion'])
        mg, zg = mZ(main_ion)
        nz = self['n_e']
        for key in self['ions_with_dens'].values:
            if key != main_ion:
                nz -= self['n_' + key].values * mZ(key)[1]
        self['n_' + main_ion] = nz / zg
        invalid = np.where(self['n_' + main_ion].values <= 0)[0]
        if len(invalid) > 0:
            printe('  Had to force main ion density to be always positive!')
            printe('  This will likely present a problem when running transport codes!')
            valid = np.where(self['n_' + main_ion].values > 0)[0]
            self['n_' + main_ion].values[invalid] = np.nanmin(self['n_' + main_ion].values[valid])


if __name__ == '__main__':

    # ensure that all specified model_tree_quantities can be translated to have <=12 chars
    for s in model_tree_species:
        for q in model_tree_quantities:
            item0 = q.format(species=s)
            item = OMFITprofiles.mds_translator(None, item0)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise


from omfit_classes.omfit_data import OMFITncDataset, DataArray, Dataset

if 'BOUT_TOP' in os.environ:
    _bout_pythonpath = os.path.abspath(os.environ['BOUT_TOP'] + os.sep + 'tools' + os.sep + 'pylib')
    if _bout_pythonpath not in [os.path.abspath(_path) for _path in sys.path]:
        sys.path.insert(1, _bout_pythonpath)
        print('* Path to BOUT python tools has been added: ' + _bout_pythonpath)
    else:
        print('* Path to BOUT python tools was found' + _bout_pythonpath)
    if 'PYTHONPATH' in os.environ and os.environ['PYTHONPATH']:
        if _bout_pythonpath not in os.environ['PYTHONPATH']:
            os.environ['PYTHONPATH'] = _bout_pythonpath + ':' + os.environ['PYTHONPATH']
    else:
        os.environ['PYTHONPATH'] = _bout_pythonpath

try:
    import netCDF4

    with quiet_environment():
        import boutdata
        import boututils
except Exception as _excp:
    boututils = boutdata = None
    warnings.warn('No `boutdata` and `boututils` support (BOUT_TOP environment variable): ' + repr(_excp))


class OMFITbout(OMFITncDataset):
    def __init__(self, filename, grid=None, tlim=None, **kw):
        if os.path.isdir(filename):
            filename = filename + os.sep + 'BOUT.dmp.0.nc'
        directory = os.path.split(filename)[0]

        # data already collected
        if not re.match(r'.*BOUT\.dmp\.[0-9]*\.nc', filename):
            OMFITncDataset.__init__(self, filename, **kw)

        # collect
        else:
            printi('Collecting BOUT++ quantities to a single file')

            tind = None
            if tlim is not None:
                t_array = boutdata.collect('t_array', path=directory, info=False)
                tind = [max([0, len(t_array) - tlim]), len(t_array)]

            OMFITncDataset.__init__(self, 'BOUT.dmp.nc')
            self.dynaLoad = False

            with netCDF4.Dataset(filename) as f:
                keys = list(f.variables.keys())
                for k, uvar in enumerate(keys):
                    var = str(uvar)
                    printi('[% 3d/% 3d] %s' % (k + 1, len(keys), var))
                    dims = list(f.variables[uvar].dimensions)
                    self[var] = DataArray(boutdata.collect(var, path=directory, info=False, tind=tind), dims=dims)

        # add grid information
        if grid is not None:
            self.add_grid(grid)

    def add_grid(self, grid):
        if grid is not None:
            self.update(grid)

    def pol_slice(self, var3d, n=1, zangle=0.0, info=True):
        """
        return 2D data from 3D array with x, y, z dimensions

        :param var3d: variable to process (string or 3D variable)

        :param n: toroidal mode number

        :param zangle: toroidal angle

        :param info: print info to screent

        :return: 2d data
        """
        n = int(n)
        zangle = float(zangle)

        if isinstance(var3d, str):
            var3d = self[var3d].values

        s = np.shape(var3d)
        if len(s) != 3:
            raise Exception('ERROR: pol_slice expects a 3D variable')

        nx, ny, nz = s

        dz = 2.0 * np.pi / float(n * (nz - 1))

        # Check the grid size is correct
        if self['nx'] != nx:
            raise Exception('ERROR: Grid X size is different to the variable')
        if self['ny'] != ny:
            raise Exception('ERROR: Grid Y size is different to the variable')

        # Get the toroidal shift
        if 'qinty' in self:
            zShift = self['qinty'].values
            if infp:
                print('Using qinty as toroidal shift angle')
        elif 'zShift' in self:
            zShift = self['zShift'].values
            if info:
                print('Using zShift as toroidal shift angle')
        else:
            raise Exception('ERROR: Neither qinty nor zShift found')

        var2d = np.zeros([nx, ny])

        ######################################
        # Perform 2D slice
        zind = (zangle - zShift) / dz
        z0f = np.floor(zind)
        z0 = z0f.astype(int)
        p = zind - z0f

        # Make z0 between 0 and (nz-2)
        z0 = ((z0 % (nz - 1)) + (nz - 1)) % (nz - 1)

        # Get z+ and z-
        zp = (z0 + 1) % (nz - 1)
        zm = (z0 - 1 + (nz - 1)) % (nz - 1)

        # There may be some more cunning way to do this indexing
        for x in np.arange(nx):
            for y in np.arange(ny):
                var2d[x, y] = (
                    0.5 * p[x, y] * (p[x, y] - 1.0) * var3d[x, y, zm[x, y]]
                    + (1.0 - p[x, y] * p[x, y]) * var3d[x, y, z0[x, y]]
                    + 0.5 * p[x, y] * (p[x, y] + 1.0) * var3d[x, y, zp[x, y]]
                )

        return var2d


__all__ = ['boutdata', 'boututils', 'OMFITbout']



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_namelist import OMFITnamelist
from omfit_classes.omfit_ascii import OMFITascii
import numpy as np
import fortranformat

__all__ = ['OMFITmhdin', 'OMFITdprobe', 'OMFITnstxMHD']


class OMFITmhdin(OMFITnamelist):
    scaleSizes = 50
    invalid = 99

    @dynaLoad
    def load(self, *args, **kw):
        r"""
        Load OMFITmhdin file

        :param \*args: arguments passed to OMFITnamelist.load()

        :param \**kw: keyword arguments passed to OMFITnamelist.load()
        """

        # load namelist
        self.outsideOfNamelistIsComment = True
        self.noSpaceIsComment = True
        super().load(*args, **kw)

        # out-of-namelist format (DIII-D)
        if 'RF' not in self['IN3']:
            fformat = {}
            fformat['FC'] = fortranformat.FortranRecordReader('(4f12.4,2A12)')
            fformat['OH'] = fortranformat.FortranRecordReader('(5f10.6)')
            fformat['VESSEL'] = fortranformat.FortranRecordReader('(4f12.4,2A12)')

            # get text past namelist
            comment_item = list(self.keys())[-1]
            if '__comment' not in comment_item:
                self.clear()
                self.dynaLoad = True
                raise ValueError('Not a valid mhdin.dat file')
            lines = str(self[comment_item])
            tmp = lines.split('comment')
            if len(tmp) > 1:
                lines, comment = tmp
            else:
                lines = tmp[0]
                comment = ''
            del self[comment_item]
            lines = lines.split('\n')
            lines = [line.expandtabs(12) for line in lines if len(line.strip()) and line.strip()[0] != '!']

            # number of elements per section
            nfc = len(self['IN3'].get('FCTURN', self['IN3'].get('TURNFC', None)))
            if self['IN3'].get('IECOIL', 0):
                nve = len(self['IN3']['RSISVS'])
                ntf = len(lines) - nfc - nve
            else:
                ntf = 0
                nve = len(lines) - nfc - ntf

            # poloidal field coils
            # R,Z,DR,DZ,skew_angle1,skew_angle2
            kk = 0
            self['FC'] = {}
            for k in range(nfc):
                lines[kk] += '%12d%12d' % (0, 0)
                data = fformat['FC'].read(lines[kk])
                data[4] = '0.0' if not data[4].strip() or float(data[4]) == 0 else data[4]
                data[5] = '90.0' if not data[5].strip() or float(data[5]) == 0 else data[5]
                data = np.array(list(map(float, data)))
                self['FC']['%d' % (k + 1)] = data
                kk += 1
            self['FC'] = np.array(self['FC'].values())

            # ohmic coils
            # R,Z,DR,DZ,block
            if not ntf:
                self['OH'] = np.array([])
            else:
                self['OH'] = {}
                for k in range(ntf):
                    data = fformat['OH'].read(lines[kk])
                    self['OH']['%d' % (k + 1)] = data
                    kk += 1
                self['OH'] = np.array(self['OH'].values())

            # conducting vessel segments
            # R,Z,DR,DZ,skew_angle1,skew_angle2
            if not nve:
                self['VESSEL'] = np.array([])
            else:
                self['VESSEL'] = {}
                for k in range(nve):
                    lines[kk] += '%12d%12d' % (0, 0)
                    data = fformat['VESSEL'].read(lines[kk])
                    data[4] = '0.0' if not data[4].strip() or float(data[4]) == 0 else data[4]
                    data[5] = '90.0' if not data[5].strip() or float(data[5]) == 0 else data[5]
                    data = np.array(list(map(float, data)))
                    data = np.array(data)
                    self['VESSEL']['%d' % (k + 1)] = data
                    kk += 1
                self['VESSEL'] = np.array(self['VESSEL'].values())

            # uncomment this to translate out-of-namelist format to in-namelist format
            if False:
                # create namelist elements
                self['IN3']['RF'] = self['FC'][:, 0]
                self['IN3']['ZF'] = self['FC'][:, 1]
                self['IN3']['WF'] = self['FC'][:, 2]
                self['IN3']['HF'] = self['FC'][:, 3]
                self['IN3']['AF'] = self['FC'][:, 4]
                self['IN3']['AF2'] = self['FC'][:, 5]

                self['IN3']['RVS'] = self['VESSEL'][:, 0]
                self['IN3']['ZVS'] = self['VESSEL'][:, 1]
                self['IN3']['WVS'] = self['VESSEL'][:, 2]
                self['IN3']['HVS'] = self['VESSEL'][:, 3]
                self['IN3']['AVS'] = self['VESSEL'][:, 5]
                self['IN3']['AVS2'] = self['VESSEL'][:, 5]

                self['IN3']['RE'] = self['OH'][:, 0]
                self['IN3']['ZE'] = self['OH'][:, 1]
                self['IN3']['WE'] = self['OH'][:, 2]
                self['IN3']['HE'] = self['OH'][:, 3]

            self['__trailing_comment__'] = 'comment' + comment

        # in-namelist format
        else:

            # initialize namelist
            for geom in ['R{element}', 'Z{element}', 'W{element}', 'H{element}', 'A{element}', 'A{element}2']:
                for element in ['E', 'F', 'VS']:
                    item = geom.format(element=element)
                    if item not in self['IN3'] and item not in ['AE', 'AE2']:
                        self['IN3'][item] = np.array([])

            # vessel and PFs are separated
            if len(self['IN3']['RVS']):
                self['FC'] = np.array(
                    [
                        self['IN3']['RF'],
                        self['IN3']['ZF'],
                        self['IN3']['WF'],
                        self['IN3']['HF'],
                        self['IN3']['AF'],
                        self['IN3']['AF2'] + (self['IN3']['AF2'] == 0) * 90,
                    ]
                ).T
                self['OH'] = np.array(
                    [self['IN3']['RE'], self['IN3']['ZE'], self['IN3']['WE'], self['IN3']['HE'], self['IN3']['HE'] * 0 + 1]
                ).T
                self['VESSEL'] = np.array(
                    [
                        self['IN3']['RVS'],
                        self['IN3']['ZVS'],
                        self['IN3']['WVS'],
                        self['IN3']['HVS'],
                        self['IN3']['AVS'],
                        self['IN3']['AVS2'] + (self['IN3']['AVS2'] == 0) * 90,
                    ]
                ).T

            # vessel as PFs
            else:
                tmp = np.unique(self['IN3']['FCID'] - 1)
                pf_ids = tmp[: -len(self['IN3']['VSNAME'])]
                vessel_ids = tmp[-len(self['IN3']['VSNAME']) :]
                split = int(np.max(pf_ids)) - 1
                PF = np.array(
                    [
                        self['IN3']['RF'],
                        self['IN3']['ZF'],
                        self['IN3']['WF'],
                        self['IN3']['HF'],
                        self['IN3']['AF'],
                        (self['IN3']['AF2'] + (self['IN3']['AF2'] == 0) * 90),
                    ]
                ).T
                self['FC'] = PF[:split, :]
                self['OH'] = np.array(
                    [self['IN3']['RE'], self['IN3']['ZE'], self['IN3']['WE'], self['IN3']['HE'], self['IN3']['HE'] * 0 + 1]
                ).T
                self['VESSEL'] = PF[split:, :]

                self['IN3']['RVS'] = self['VESSEL'][:, 0]
                self['IN3']['ZVS'] = self['VESSEL'][:, 1]
                self['IN3']['WVS'] = self['VESSEL'][:, 2]
                self['IN3']['HVS'] = self['VESSEL'][:, 3]
                self['IN3']['AVS'] = self['VESSEL'][:, 4]
                self['IN3']['AVS2'] = self['VESSEL'][:, 5]

    @dynaSave
    def save(self, *args, **kw):
        r"""
        Save OMFITmhdin file

        :param \*args: arguments passed to OMFITnamelist.save()

        :param \**kw: keyword arguments passed to OMFITnamelist.save()
        """
        fformat = {}
        fformat['FC'] = fortranformat.FortranRecordWriter('(4f12.4,2A12)')
        fformat['OH'] = fortranformat.FortranRecordWriter('(5f10.6)')
        fformat['VESSEL'] = fortranformat.FortranRecordWriter('(4f12.4,2A12)')

        mhdin_special = ['FC', 'OH', 'VESSEL', '__trailing_comment__']

        # remove non-namelist components
        tmp = {}
        for item in mhdin_special:
            if item in self:
                tmp[item] = self[item]
                del self[item]
        angle2_special = {}
        for item in ['AF2', 'AVS2']:
            if item not in self['IN3']:
                continue
            angle2_special[item] = self['IN3'][item].copy()
            self['IN3'][item][self['IN3'][item] == 90] = 0.0
        try:
            # save namelist section
            super().save(*args, **kw)

            # append non-namelist components
            txt = ''
            for system in mhdin_special:
                if system not in tmp:
                    continue
                if system in fformat:
                    for item in range(tmp[system].shape[0]):
                        data = tmp[system][item, :].tolist()
                        if len(data) == 6:
                            if data[4] == 0:
                                data[4] = ''
                            if data[5] == 90:
                                data[5] = ''
                        txt += fformat[system].write(data) + '\n'
                else:
                    txt += tmp[system]
            with open(self.filename, 'a') as f:
                f.write(txt)

        finally:
            # restore angles
            for item in angle2_special:
                self['IN3'][item] = angle2_special
            # restore non-namelist components
            for item in mhdin_special:
                if item in tmp:
                    self[item] = tmp[item]

    @staticmethod
    def plot_coil(data, patch_facecolor='lightgray', patch_edgecolor='black', label=None, ax=None):
        """
        plot individual coil

        :param data: FC, OH, VESSEL data array row

        :param patch_facecolor: face color

        :param patch_edgecolor: edge color

        :param label: [True, False]

        :param ax: axis

        :return: matplotlib rectangle patch
        """
        import matplotlib.transforms as mtransforms
        from matplotlib import patches

        if ax is None:
            ax = pyplot.gca()

        rect = patches.Rectangle((0, 0), data[2], data[3], facecolor=patch_facecolor, edgecolor=patch_edgecolor)
        if len(data) == 6:
            angle1, angle2 = 90 - data[5], data[4]
            rect.set_transform(
                mtransforms.Affine2D().translate(-data[2] / 2.0, -data[3] / 2.0)
                + mtransforms.Affine2D().skew_deg(angle1, angle2)
                + mtransforms.Affine2D().translate(data[0], data[1])
                + ax.transData
            )
        else:
            rect.set_transform(
                mtransforms.Affine2D().translate(-data[2] / 2.0, -data[3] / 2.0)
                + mtransforms.Affine2D().translate(data[0], data[1])
                + ax.transData
            )
        ax.add_patch(rect)

        if label:
            ax.text(data[0], data[1], label, color='w', size=8, ha='center', va='center', zorder=1000, weight='bold', clip_on=True)
            ax.text(data[0], data[1], label, color='m', size=8, ha='center', va='center', zorder=1001, clip_on=True)
        return rect

    def plot_flux_loops(self, display=None, colors=None, label=False, ax=None):
        """
        plot the flux loops

        :param display: array used to turn on/off display individual flux loops

        :param colors: array used to set the color of individual flux loops

        :param label: [True, False]

        :param ax: axis
        """
        if 'ZSI' not in self['IN3'] or 'RSI' not in self['IN3']:
            return
        if ax is None:
            ax = pyplot.gca()
        x0 = np.squeeze(self['IN3']['RSI'])
        y0 = np.squeeze(self['IN3']['ZSI'])
        if colors is not None:
            c0 = np.squeeze(colors)[: len(x0)]
        if display is not None:
            s0 = np.squeeze((display != 0))[: len(x0)]
        else:
            s0 = np.ones(x0.shape)
        s0 *= self.scaleSizes

        # trim
        x0 = x0[: len(s0)]
        y0 = y0[: len(s0)]

        # disable plotting of dummy flux loops
        x0 = x0[np.where(y0 != -self.invalid)]
        y0 = y0[np.where(y0 != -self.invalid)]

        # plot
        if colors is not None:
            ax.scatter(x0, y0, s=s0, c=c0, vmin=0, vmax=vmax, marker='o', cmap=cm, alpha=0.75, zorder=100)
        else:
            ax.scatter(x0, y0, s=s0, color='b', marker='o', alpha=0.75, zorder=100)

        # labels
        if label:
            for k, name in enumerate(self['IN3']['LPNAME']):
                ax.text(x0[k], y0[k], '\n ' + name, ha='left', va='top', size=8, color='w', zorder=1000, weight='bold', clip_on=True)
                ax.text(x0[k], y0[k], '\n ' + name, ha='left', va='top', size=8, color='b', zorder=1001, clip_on=True)

    def plot_magnetic_probes(self, display=None, colors=None, label=False, ax=None):
        """
        plot the magnetic probes

        :param display: array used to turn on/off the display of individual magnetic probes

        :param colors: array used to set the color of individual magnetic probes

        :param label: [True, False]

        :param ax: axis
        """
        # The magnetic probes are characterized by:
        #  - XMP2 and
        #  - YMP2, cartesian coordinates of the center of the probe,
        #  - SMP2, size/length of the probe in meters (read below!),
        #  - AMP2, angle/orientation of the probe in degrees.
        #
        # the usual magnetic probe in EFIT is a partial rogowski coil,
        # yet beware! the EFIT D3D probe file also models saddle loops,
        # which extend in the toroidal direction and provide integrated
        # signals. such loops are characterized by a negative length.
        #
        # in order to plot non-rogowski coils correctly, a forced 90 deg
        # counter-clockwise rotation has to be applied on the probe's angle.
        #
        # the probes are plotted with different linestyles: rogowski coils
        # are plotted with a segment centered around a dot, whereas
        # saddle loops are plotted with a segment with dots on the endpoints.
        #
        # FURTHER REFERENCE as explained by T. Strait on 19-jul-2016
        #
        # - The angle AMP2 always indicates the direction of the magnetic field
        #   component that is being measured.
        #
        # - The length SMP2 indicates the length (in the R-Z plane) over which
        #   the magnetic field is averaged by the sensor.
        #
        # - SMP2 > 0 indicates that the averaging length is in the direction of AMP2.
        #   SMP2 < 0 indicates that the averaging length is perpendicular to AMP2.
        #
        # - In predicting the measurement of the sensor for purposes of fitting,
        #   only the length SMP2 is considered.  The width of the sensor in the
        #   direction perpendicular to SMP2 (in the R-Z plane) is small and is
        #   therefore neglected.
        #   Since the EFIT equilibrium is assumed to be axisymmetric, the width
        #   of the sensor in the toroidal direction is not relevant.
        #
        if 'XMP2' not in self['IN3'] or 'YMP2' not in self['IN3'] or 'SMP2' not in self['IN3'] or 'AMP2' not in self['IN3']:
            return
        if ax is None:
            ax = pyplot.gca()
        # first, get the arrays and make sure that their dimensions match
        x0 = np.squeeze(self['IN3']['XMP2'])
        y0 = np.squeeze(self['IN3']['YMP2'])
        l0 = np.squeeze(self['IN3']['SMP2'])
        a0 = np.squeeze(self['IN3']['AMP2'])
        if colors is not None:
            c0 = np.squeeze(colors)[: len(x0)]
        if display is not None:
            s0 = np.squeeze((display != 0))[: len(x0)]
        else:
            s0 = np.ones(x0.shape)
        s0 *= self.scaleSizes

        # trim
        x0 = x0[: len(s0)]
        y0 = y0[: len(s0)]
        l0 = l0[: len(s0)]
        a0 = a0[: len(s0)]

        # disable plotting of dummy probes
        l0 = l0[np.where(y0 != -self.invalid)]
        a0 = a0[np.where(y0 != -self.invalid)]
        x0 = x0[np.where(y0 != -self.invalid)]
        y0 = y0[np.where(y0 != -self.invalid)]

        def probe_endpoints(x0, y0, a0, l0):
            boo = (1 - np.sign(l0)) / 2.0
            cor = boo * np.pi / 2.0

            # then, compute the two-point arrays to build the partial rogowskis
            # as segments rather than single points, applying the correction
            px = x0 - l0 / 2.0 * np.cos(a0 * np.pi / 180.0 + cor)
            py = y0 - l0 / 2.0 * np.sin(a0 * np.pi / 180.0 + cor)
            qx = x0 + l0 / 2.0 * np.cos(a0 * np.pi / 180.0 + cor)
            qy = y0 + l0 / 2.0 * np.sin(a0 * np.pi / 180.0 + cor)

            segx = []
            segy = []
            for k in range(len(x0)):
                segx.append([px[k], qx[k]])
                segy.append([py[k], qy[k]])
            return segx, segy

        # finally, plot
        segx, segy = probe_endpoints(x0, y0, a0, l0)
        for k in range(len(x0)):
            if colors is None:
                col = 'r'
            else:
                col = cm(c0[k])
            if l0[k] > 0:
                ax.plot(segx[k], segy[k], '-', lw=2, color=col, zorder=100, alpha=0.75)
                ax.plot(x0[k], y0[k], '.', color=col, zorder=100, alpha=0.75, mec='none')
            else:
                ax.plot(segx[k], segy[k], '.-', lw=2, color=col, zorder=100, alpha=0.75, mec='none')

        # labels
        if label:
            for k, name in enumerate(self['IN3']['MPNAM2']):
                ax.text(x0[k], y0[k], '\n ' + name, ha='left', va='top', size=8, color='w', zorder=1000, weight='bold', clip_on=True)
                ax.text(x0[k], y0[k], '\n ' + name, ha='left', va='top', size=8, color='r', zorder=1001, clip_on=True)

    def plot_poloidal_field_coils(self, edgecolor='none', facecolor='orange', label=False, ax=None):
        """
        Plot poloidal field coils

        :param label: [True, False]

        :param ax: axis
        """
        if not self['FC'].size:
            return
        if ax is None:
            ax = pyplot.gca()
        return self.plot_system('FC', edgecolor=edgecolor, facecolor=facecolor, label=label, ax=ax)

    def plot_ohmic_coils(self, edgecolor='none', facecolor='none', label=False, ax=None):
        """
        Plot ohmic coils

        :param label: [True, False]

        :param ax: axis
        """
        if not self['OH'].size:
            return
        if ax is None:
            ax = pyplot.gca()
        return self.plot_system('OH', edgecolor=edgecolor, facecolor=facecolor, label=label, ax=ax)

    def plot_vessel(self, edgecolor='none', facecolor='gray', label=False, ax=None):
        """
        Plot vacuum vessel

        :param label: [True, False]

        :param ax: axis
        """
        if not self['VESSEL'].size:
            return
        if ax is None:
            ax = pyplot.gca()
        return self.plot_system('VESSEL', edgecolor=edgecolor, facecolor=facecolor, label=label, ax=ax)

    def plot_system(self, system, edgecolor, facecolor, label=False, ax=None):
        """
        Plot coil/vessel system

        :param system: ['FC', 'OH', 'VESSEL']

        :param edgecolor: color of patch edges

        :param facecolor: color of patch fill

        :param label: [True, False]

        :param ax: axis
        """
        if ax is None:
            ax = pyplot.gca()
        kw = {'ax': ax}
        kw['patch_facecolor'] = facecolor
        kw['patch_edgecolor'] = edgecolor
        if system == 'OH':
            n = int(max(self[system][:, -1]))
        for k in range(self[system].shape[0]):
            # disable plotting of dummy probes/loops
            if self[system][k, 1] == self.invalid:
                continue
            if system == 'OH' and facecolor == 'none':
                kw['patch_facecolor'] = pyplot.cm.viridis(np.linspace(0.0, 1.0, n))[int(self[system][k, -1]) - 1]
            if label:
                kw['label'] = '%d' % k
            self.plot_coil(self[system][k, :], **kw)
        ax.set_frame_on(False)
        ax.set_aspect('equal')
        ax.autoscale(tight=True)
        ax.set_xlim(ax.get_xlim() * np.array([0.98, 1.02]))
        ax.set_ylim(ax.get_ylim() * np.array([1.02, 1.02]))

    def plot_domain(self, ax=None):
        """
        plot EFUND computation domain

        :param ax: axis
        """
        if ax is None:
            ax = pyplot.gca()

        from matplotlib import patches

        rect = patches.Rectangle(
            (self['IN3']['RLEFT'], self['IN3']['ZBOTTO']),
            self['IN3']['RRIGHT'] - self['IN3']['RLEFT'],
            self['IN3']['ZTOP'] - self['IN3']['ZBOTTO'],
            facecolor='none',
            edgecolor='black',
            ls='--',
        )
        ax.add_patch(rect)

    def plot(self, label=False, plot_coils=True, plot_vessel=True, plot_measurements=True, plot_domain=True, ax=None):
        """
        Composite plot

        :param label: label coils and measurements

        :param plot_coils: plot poloidal field and oh coils

        :param plot_vessel: plot conducting vessel

        :param plot_measurements: plot flux loops and magnetic probes

        :param plot_domain: plot EFUND computing domain

        :param ax: axis
        """

        if ax is None:
            ax = pyplot.gca()
        if plot_coils:
            self.plot_poloidal_field_coils(label=label, ax=ax)
            self.plot_ohmic_coils(label=label, ax=ax)
        if plot_vessel:
            self.plot_vessel(label=label, ax=ax)
        if plot_measurements:
            self.plot_flux_loops(label=label, ax=ax)
            self.plot_magnetic_probes(label=label, ax=ax)
        if not isinstance(self, OMFITdprobe) and plot_domain:
            self.plot_domain(ax=ax)
        ax.autoscale(tight=True)

    def __call__(self, *args, **kw):
        r"""
        Done to override default OMFIT GUI behaviour for OMFITnamelist
        """
        return self.plotFigure(*args, **kw)

    def aggregate_oh_coils(self, index=None, group=None):
        """
        Aggregate selected OH coils into a single coil

        :param index of OH coils to aggregate

        :param group: group of OH coils to aggregate
        """
        if group is not None:
            index = np.where(self['OH'][:, 4] == group)[0]

        mx = np.min(self['OH'][index, 0] - self['OH'][index, 2] / 2.0)
        MX = np.max(self['OH'][index, 0] + self['OH'][index, 2] / 2.0)
        my = np.min(self['OH'][index, 1] - self['OH'][index, 3] / 2.0)
        MY = np.max(self['OH'][index, 1] + self['OH'][index, 3] / 2.0)
        group = self['OH'][index, 4][0]

        aggregated_coil = [(MX + mx) / 2.0, (MY + my) / 2.0, (MX - mx), (MY - my), group]

        index = np.array(sorted(list(set(list(range(self['OH'].shape[0]))).difference(list(index)))))
        self['OH'] = self['OH'][index, :]
        self['OH'] = np.vstack((self['OH'], aggregated_coil))

    def disable_oh_group(self, group):
        """
        remove OH group

        :param group: group of OH coils to disable
        """
        index = np.where(self['OH'][:, 4] != group)[0]
        self['OH'] = self['OH'][index, :]
        groups = np.unique(self['OH'][:, 4])
        groups_mapper = dict(zip(groups, range(1, len(groups) + 1)))
        for k in range(self['OH'].shape[0]):
            self['OH'][k, 4] = groups_mapper[self['OH'][k, 4]]

    def change_R(self, deltaR=0.0):
        """
        add or subtract a deltaR to coils, flux loops and magnetic
        probes radial location effectively changing the aspect ratio

        :param deltaR: radial shift in m
        """
        for system in ['FC', 'OH', 'VESSEL']:
            self[system][:, 0] += deltaR
        for item in self['IN3']:
            if item[0].upper() in ['R', 'X']:
                self['IN3'][item] += deltaR

    def change_Z(self, deltaZ=0.0):
        """
        add or subtract a deltaZ to coils, flux loops and magnetic
        probes radial location effectively changing the aspect ratio

        :param deltaR: radial shift in m
        """
        for system in ['FC', 'OH', 'VESSEL']:
            self[system][:, 1] += deltaZ
        for item in self['IN3']:
            if item[0].upper() in ['Z', 'Y']:
                self['IN3'][item] += deltaZ

    def scale_system(self, scale_factor=0.0):
        """
        scale coils, flux loops and magnetic
        probes radial location effectively changing the major radius
        holding aspect ratio fixed

        :param scale_factor: scaling factor to multiple system by
        """
        for system in ['FC', 'OH', 'VESSEL']:
            self[system][:, 0] *= scale_factor
            self[system][:, 1] *= scale_factor
        for item in self['IN3']:
            if item[0].upper() in ['R', 'X']:
                self['IN3'][item] *= scale_factor
            if item[0].upper() in ['Z', 'Y']:
                self['IN3'][item] *= scale_factor

    def fill_coils_from(self, mhdin):
        """
        Copy FC, OH, VESSEL from passed object into current object,
        without changing the number of elements in current object.

        This requires that the number of elements in the current object
        is greater or equal than the number of elements in the passed object.
        The extra elements in the current object will be placed at R=0, Z=0

        :param mhdin: other mhdin object
        """
        self['FC'][:, 0] = 0.01
        self['FC'][:, 1] = self.invalid
        self['FC'][:, 2] = 0.01
        self['FC'][:, 3] = 0.01
        self['FC'][:, 4] = 0.0
        self['FC'][:, 5] = 90.0

        self['OH'][:, 0] = 0.01
        self['OH'][:, 1] = self.invalid
        self['OH'][:, 2] = 0.01
        self['OH'][:, 3] = 0.01
        if len(mhdin['OH']):
            delta_shape = self['OH'].shape[0] - mhdin['OH'].shape[0]
            if delta_shape and max(mhdin['OH'][:, 4]) >= max(self['OH'][:, 4]):
                raise ValueError('OMFITmhdin.fill_coils_from() has no space for an extra `invalid` OH group')
            self['OH'][mhdin['OH'].shape[0] :, 4] = np.linspace(
                max(mhdin['OH'][:, 4]) + 1, max(self['OH'][:, 4]) + 0.9999, delta_shape
            ).astype(int)
            self['OH'][: mhdin['OH'].shape[0], 4] = 0.0

        self['VESSEL'][:, 0] = 0.01
        self['VESSEL'][:, 1] = self.invalid
        self['VESSEL'][:, 2] = 0.01
        self['VESSEL'][:, 3] = 0.01
        self['VESSEL'][:, 4] = 0.0
        self['VESSEL'][:, 5] = 90.0
        self['IN3']['VSNAME'] = ['DUMMY_%d' % k for k in range(self['VESSEL'].shape[0])]
        if len(mhdin['VESSEL']):
            self['IN3']['VSNAME'][: self['VESSEL'].shape[0]] = mhdin['IN3'].get('VSNAME', ['vessel'] * self['VESSEL'].shape[0])

        for system in ['FC', 'OH', 'VESSEL']:
            if len(mhdin[system]):
                self[system][: len(mhdin[system])] = mhdin[system]

    def modify_vessel_elements(self, index, action=['keep', 'delete'][0]):
        """
        Utility function to remove vessel elements

        :param index: index of the vessel elements to either keep or delete

        :param action: can be either 'keep' or 'delete'
        """
        keep_index = index
        if action == 'delete':
            keep_index = [k for k in range(len(self['IN3']['VSNAME'])) if k not in index]
        self['VESSEL'] = self['VESSEL'][keep_index]
        self['IN3']['VSNAME'] = np.array(self['IN3']['VSNAME'])[keep_index]

    def fill_probes_loops_from(self, mhdin):
        """
        Copy flux loops and magnetic probes from other object into current object,
        without changing the number of elements in current object

        This requires that the number of elements in the current object
        is greater or equal than the number of elements in the passed object.
        The extra elements in the current object will be placed at R=0, Z=0

        :param mhdin: other mhdin object
        """
        for system in [['XMP2', 'YMP2', 'SMP2', 'AMP2', 'MPNAM2'], ['RSI', 'ZSI', 'LPNAME']]:
            for item in system:
                if item in ['MPNAM2', 'LPNAME']:
                    self['IN3'][item] = ['DUMMY_%d' % k for k in range(len(self['IN3'][item]))]
                    self['IN3'][item][: len(mhdin['IN3'][item])] = mhdin['IN3'][item]
                else:
                    self['IN3'][item] *= 0
                    if item[0] in ['X', 'R']:
                        self['IN3'][item] += 0.01
                    elif item[0] in ['Y', 'Z']:
                        self['IN3'][item] += -self.invalid
                    elif item[0] in ['S']:
                        self['IN3'][item] += 0.01
                    self['IN3'][item][: len(mhdin['IN3'][item])] = mhdin['IN3'][item]

    def fill_scalars_from(self, mhdin):
        """
        copy scalar quantities in IN3 namelist
        without overwriting ['IFCOIL', 'IECOIL', 'IVESEL']

        :param mhdin: other mhdin object
        """
        for item in self['IN3']:
            if item in mhdin['IN3']:
                if isinstance(mhdin['IN3'][item], (int, float)):
                    if item not in ['IFCOIL', 'IECOIL', 'IVESEL']:
                        self['IN3'][item] = mhdin['IN3'][item]
                else:
                    self['IN3'][item] = np.array(self['IN3'][item])

    def pretty_print(self, default_tilt2=0):
        for item in ['FC', 'OH', 'VESSEL']:
            print('# ' + '=' * len(item))
            print('# ' + item)
            print('# ' + '=' * len(item))
            for row in range(self[item].shape[0]):
                start = ' '
                end = ','
                if row == 0:
                    start = '['
                if row == self[item].shape[0] - 1:
                    end = ']'
                tmp = self[item][row, :].copy()
                if len(tmp) == 6 and tmp[5] == 90:
                    tmp[5] = default_tilt2
                print(start + '[' + ', '.join(map(lambda x: repr(x), tmp)) + ']' + end)

        print('# ==========')
        print('# Flux loops')
        print('# ==========')
        print('R_flux_loop = ', end='')
        print(repr(self['IN3']['RSI']))
        print('Z_flux_loop = ', end='')
        print(repr(self['IN3']['ZSI']))
        print('name_flux_loop = ', end='')
        print(repr(self['IN3']['LPNAME']))

        print('# ===============')
        print('# Magnetic probes')
        print('# ===============')
        print('R_magnetic = ', end='')
        print(repr(self['IN3']['XMP2']))
        print('Z_magnetic = ', end='')
        print(repr(self['IN3']['YMP2']))
        print('A_magnetic = ', end='')
        print(repr(self['IN3']['AMP2']))
        print('S_magnetic = ', end='')
        print(repr(self['IN3']['SMP2']))
        print('name_magnetic = ', end='')
        print(repr(self['IN3']['MPNAM2']))

        return self


class OMFITdprobe(OMFITmhdin):
    @dynaLoad
    def load(self, *args, **kw):
        super().load(*args, **kw)
        for item in list(self['IN3'].keys()):
            if item.upper() not in [
                'RSISVS',
                'TURNFC',
                'VSNAME',
                'LPNAME',
                'MPNAM2',
                'RSI',
                'ZSI',
                'XMP2',
                'YMP2',
                'AMP2',
                'SMP2',
                'PATMP2',
                'IECOIL',
                'IVESEL',
                'IFCOIL',
            ]:
                del self['IN3'][item]


class OMFITnstxMHD(SortedDict, OMFITascii):
    """
    OMFIT class to read NSTX MHD device files such as `device01152015.dat`, `diagSpec01152015.dat` and `signals_020916_PF4.dat`
    """

    def __init__(self, filename, use_leading_comma=None, **kw):
        r'''
        OMFIT class to parse NSTX MHD device files

        :param filename: filename

        :param \**kw: arguments passed to __init__ of OMFITascii
        '''
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.read().split('\n')
        lines = [line for line in lines if len(line.strip()) and not line.startswith(';')]
        for line in lines:
            line, description = (line + ';').split(';', 1)
            line = line.split()
            if line[1] not in self:
                self[line[1]] = {}
            piece = len(self[line[1]])
            # signals.dat
            if len(line) == 17:
                self.type = 'signals'
                self[line[1]][piece] = dict(
                    zip(
                        'name type map_var map_index mds_tree read_sig rel_error abs_error sig_thresh use_err scale pri fitwt t_index t_smooth mds_name'.split(),
                        line,
                    )
                )
            # diagspec.dat
            elif len(line) == 12:
                self.type = 'diagspec'
                self[line[1]][piece] = dict(zip('name type id rc zc wc hc ac ac2 turns div material'.split(), line))
            # device.dat
            elif len(line) == 8:
                self.type = 'device'
                self[line[1]][piece] = dict(zip('tag type units rc zc pol_ang tor_ang1 tor_ang2'.split(), line))
            else:
                raise ValueError(f'Format not recognized {len(line)}')
            self[line[1]][piece]['description'] = description.strip(';')
            for item in self[line[1]][piece]:
                try:
                    self[line[1]][piece][item] = ast.literal_eval(self[line[1]][piece][item])
                except (ValueError, SyntaxError):
                    pass

        # postprocess
        if self.type == 'signals':
            mds_tree_map = {'o': 'operations', 'e': 'engineering', 'a': 'activespec', 'r': 'radiatiion', 'E2': 'efit02', '-': 'computed'}
            for group in self:
                for item in self[group]:
                    self[group][item]['mds_tree'] = mds_tree_map[self[group][item]['mds_tree']]

        return self

    def pretty_print(self):
        '''
        Print data in file as arrays, as it is needed for a fortran namelist
        '''
        for group in self:
            for item in self[group][0]:
                print(f"{group.lower()}_{item.split('(')[0].lower()} = ", end='')
                print([self[group][k][item] for k in self[group]])
        return self


############################################
if '__main__' == __name__:
    from matplotlib import pyplot

    test_classes_main_header()
    for machine in ['nstx-u', 'diii-d']:
        tmp = OMFITmhdin(OMFITsrc + f'/../modules/EFUND/TEMPLATES/mhdin_{machine}.dat')
        tmp.load()
        tmp.pretty_print()
        tmp.plot(label=True)
        pyplot.show()
